"""Background tasks for QMI.

This module provides support for tasks which run asynchronously in the background.

Defining tasks
##############

To define a new background task, create a subclass of `QMI_Task`.
The new class must define a `run()` method, containing the main code of the task.
It may also be helpful to redefine the `__init__()` method in order
to accept arguments and initialize the task.

A task may define two types of special variables:

- **settings**: data that affects the behaviour of the task.
  Settings can be modified by other tasks or scripts while the task
  is still running.

- **status**: data that represents the current activity of the task.
  The task can change its own status while it runs, but other tasks or
  scripts can not change this task's status.

Declare two custom named tuple types to contain the settings resp. the status
of your task. Instances of these named tuples will be used to store the
current settings in ``self.settings`` and the current status in ``self.status``.

Example of a custom task definition::

    MySettings = namedtuple("MySettings", "param other_param")
    MyStatus = namedtuple("MyStatus", "field other_field")

    class MyTask(QMI_Task):
        def __init__(self, task_runner, name, my_arg, my_other_arg):
            super().__init__(task_runner, name)
            self.settings = MySettings(...)
            self.status = MyStatus(...)
            ...
        def run(self):
            while not self.stop_requested():
                self.update_settings()
                ...

Running tasks
#############

Tasks are created via the QMI context.
The context then returns a `proxy` for the new task instance.
A newly created task must be explicitly started by calling the `start()`
method through the proxy.

A task stops when its `run()` method returns.
Alternatively, the `stop()` method may be called through the proxy to request
that the task stops itself as soon as possible.
Even in this case, the task will still continue to run until its `run()` method
honours the stop request by returning.

To clean up a stopped task, call the `join()` method through the proxy.
The `join()` method will wait until the task has stopped, if necessary,
then clean up the background thread.

Example of starting and stopping a task::

    proxy = qmi.make_task("my_task", MyTask, "arg1", "arg2")
    proxy.start()
    ...
    new_settings = MySettings(...)
    proxy.set_settings(new_settings)
    ...
    status = proxy.get_status()
    print("Status:", status)
    ...
    proxy.stop()
    proxy.join()

Custom Task Runner
##################

It's possible to define a custom task runner and have qmi use _it_ instead of a
default `QMI_TaskRunner` instance. This will allow custom RPC methods and state
to be added to the runner, which can be useful in some cases.

Example of a custom task runner::

    class MyTaskRunner(QMI_TaskRunner):
        @rpc_method
        def set_param(self, param):
            settings = self.get_pending_settings()
            settings._replace(param=param)
            self.set_settings(settings)

    proxy = qmi.make_task("my_task", MyTask, "arg1", "arg2", task_runner=MyTaskRunner)
    proxy.start()
    ...
    proxy.set_param(...)
    ...
    proxy.stop()
    proxy.join()

Reference
#########
"""

import collections
import enum
import inspect
import logging
import threading
import typing
from typing import Callable, Optional, Tuple, Type, Any, TypeVar, Generic
import time

from qmi.core.rpc import QMI_RpcObject, rpc_method
from qmi.core.pubsub import SignalDescription, QMI_Signal, QMI_RegisteredSignal
from qmi.core.thread import QMI_Thread
from qmi.core.exceptions import (
    QMI_TaskInitException, QMI_TaskRunException, QMI_TaskStopException,
    QMI_UsageException, QMI_WrongThreadException)
from qmi.core.util import is_valid_object_name

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

# Global definitions for type variables used in this module
_SET = TypeVar("_SET")  # task settings
_STS = TypeVar("_STS")  # task status


class QMI_LoopTaskMissedLoopPolicy(enum.Enum):
    """Possible policies of QMI_LoopTask instance missing a loop period."""
    IMMEDIATE = 1  # Do actions as quickly as possible
    SKIP = 2  # Skip actions and try again at the next period
    TERMINATE = 3  # Not acceptable to miss periodical loop instances -> terminate QMI_LoopTask


class _TaskMetaClass(type):
    """Meta-class used to create QMI_Task and its subclasses.

    This meta-class extracts a list of signals published by the class.
    The list of signals is inserted as a class attribute named `_qmi_signals`.
    """

    def __init__(cls, name: str, bases: tuple, dct: dict) -> None:

        # Let "type" do its thing.
        super().__init__(name, bases, dct)

        # The original class must not yet have a _qmi_signals attribute.
        assert "_qmi_signals" not in dct

        # Scan the newly created QMI_Task subclass for class attributes of type QMI_Signal.
        # These are markers to announce signals that the class may publish.
        signals = []
        for attr_name, attr_value in inspect.getmembers(cls, lambda member: isinstance(member, QMI_Signal)):
            if not is_valid_object_name(attr_name):
                raise QMI_UsageException("Invalid signal name {!r} in class {}".format(attr_name, cls.__name__))
            signals.append(SignalDescription(name=attr_name, arg_types=attr_value.arg_types))

        # Add the list of signals as an attribute of the newly created QMI_Task subclass.
        cls._qmi_signals = tuple(signals)


class QMI_Task(Generic[_SET, _STS], metaclass=_TaskMetaClass):
    """Base class for all background tasks.

    Each instance of `QMI_Task` runs in a separate thread where it performs
    background activities without blocking the rest of the application.

    Subclasses of `QMI_Task` implement specific background tasks.
    Each subclass must implement a special method `run()`, containing
    the code which runs in the background.

    Tasks may publish QMI signals.

    Attributes:
        sig_settings_updated: Signal used to publish the latest settings when updated. Can be used by clients who need
                              to know when the settings might have changed by another client. Users need to re-define
                              this signal with appropriate type, if necessary.
    """

    sig_settings_updated = QMI_Signal([type(None)])

    def __init__(self, task_runner: 'QMI_TaskRunner', name: str) -> None:
        """Initialize the task.

        Subclasses of QMI_Task may override this method. The `__init__()`
        function of the subclass must also take `task_runner` and `name`
        as its first two parameters. These parameters are provided by
        the QMI context. The subclass `__init__()` must call the base class
        method via ``super().__init__(task_runner, name)``.

        The subclass `__init__()` may take additional parameters. These
        additional parameters can be passed to `context.make_task()`.

        Parameters:
            task_runner: The `QMI_TaskRunner` instance which manages this task.
            name: The RPC object name of this task instance.
        """

        # Check that task is initialized inside the task thread.
        if task_runner._thread is not threading.current_thread():
            raise QMI_WrongThreadException()

        _logger.debug("Initializing task %s", name)

        self._context = task_runner._context
        self._task_runner = task_runner
        self._name = name
        self._stop_requested = threading.Event()
        self._settings_fifo: collections.deque = collections.deque(maxlen=1)
        self.settings: Optional[_SET] = None
        self.status: Optional[_STS] = None

        # Create instances of QMI_RegisteredSignal for each signal type that this class may publish.
        # Insert these QMI_RegisteredSignal instances as attributes of the RpcObject instance.
        #
        # Note: The class attribute "_qmi_signals" will be created by the metaclass.
        declared_signals = self._qmi_signals  # type: ignore  # pylint: disable=no-member
        for sig_desc in declared_signals:
            if sig_desc.name == "sig_settings_updated":
                # specifically remap settings because this is only typed in the inherited task
                sig = QMI_RegisteredSignal(self._context, name, sig_desc.name, tuple([type(self.settings)]))
            else:
                sig = QMI_RegisteredSignal(self._context, name, sig_desc.name, sig_desc.arg_types)
            setattr(self, sig_desc.name, sig)

    def stop_requested(self) -> bool:
        """Return True if the task should stop.

        A long running background task should regularly call this function
        to check whether a stop request has been sent to the task.

        If this function returns True, the task should stop as
        soon as possible (by returning from its run() function).
        """
        return self._stop_requested.is_set()

    def sleep(self, duration: float) -> None:
        """Sleep for the specified duration (in seconds).

        This should be used by the task code instead of `time.sleep()`.
        If the task receives a stop request while sleeping,
        this method ends immediately and raises `QMI_TaskStopException`.

        Parameters:
            duration: Sleep duration in seconds.

        Raises:
            ~qmi.core.exceptions.QMI_TaskStopException: If the task received a stop request.
        """

        # Check that this method is called only from within the task thread.
        if self._task_runner._thread is not threading.current_thread():
            raise QMI_WrongThreadException()

        if self._stop_requested.wait(duration):
            raise QMI_TaskStopException()

    def update_settings(self) -> bool:
        """Update to the latest settings that were sent to the task.

        If another routine called `set_settings()` to send new settings
        to this task, this method copies the new settings to `self.settings`
        and returns True.

        Otherwise, the settings remain the same and this method returns False.

        Returns:
            True if there are new settings, False if the settings are unchanged.
        """

        # Check that this method is called only from within the task thread.
        if self._task_runner._thread is not threading.current_thread():
            raise QMI_WrongThreadException()

        # Try to obtain new settings.
        if self._settings_fifo:
            self.settings = self._settings_fifo.pop()
            self.sig_settings_updated.publish(self.settings)
            return True

        return False

    def run(self) -> None:
        """Main function of the task.

        This function runs in the background as long as the task is active.
        Subclasses must provide their own implementation.
        """
        raise NotImplementedError()


class _TaskThread(QMI_Thread):
    """A TaskThread is a dedicated thread which runs a QMI task.

    This is an internal QMI class, used exclusively by the QMI_TaskRunner.
    This class should never be used or accessed directly in application code.
    """

    class State(enum.Enum):
        """State of the task thread."""
        INITIAL = 100
        EXCEPTION_WHILE_INSTANTIATING_TASK = 101
        READY_TO_RUN = 102
        RUNNING = 103
        EXCEPTION_WHILE_RUNNING_TASK = 104
        TASK_COMPLETED_NORMALLY = 105
        TASK_STOPPED_BEFORE_START = 106

    def __init__(self,
                 task_runner: 'QMI_TaskRunner',
                 task_name: str,
                 task_class: Type[QMI_Task],
                 task_args: tuple,
                 task_kwargs: dict
                 ) -> None:
        super().__init__()
        self._task_runner = task_runner
        self._task_name = task_name
        self._task_class = task_class
        self._task_args = task_args
        self._task_kwargs = task_kwargs
        self.task: Optional[QMI_Task] = None
        self._exception: Optional[BaseException] = None
        self._state = _TaskThread.State.INITIAL
        self._state_cond = threading.Condition()
        self._wait_cond_lock = threading.Lock()
        self._wait_cond: Optional[threading.Condition] = None

    def run(self) -> None:
        """Main function inside the thread."""

        try:
            # Create the task instance.
            self.task = self._task_class(self._task_runner,  # type: ignore
                                         self._task_name,
                                         *self._task_args,
                                         **self._task_kwargs)
        except BaseException as exception:
            # Initialization failed. Report exception and stop the thread.
            with self._state_cond:
                _logger.warning("Initialization of task %s failed", self._task_name, exc_info=True)
                self._exception = exception
                self._state = _TaskThread.State.EXCEPTION_WHILE_INSTANTIATING_TASK
                self._state_cond.notify_all()
                return

        # Notify the outside world that the task was initialized.
        with self._state_cond:
            assert self._state in (_TaskThread.State.INITIAL, _TaskThread.State.TASK_STOPPED_BEFORE_START)
            if self._state == _TaskThread.State.INITIAL:
                self._state = _TaskThread.State.READY_TO_RUN
                self._state_cond.notify_all()

        _logger.debug("Task thread %s ready to run", self._task_name)

        # Wait until the outside world tells us to continue.
        with self._state_cond:
            while self._state == _TaskThread.State.READY_TO_RUN:
                self._state_cond.wait()
            if self._state != _TaskThread.State.RUNNING:
                # The task was stopped before it was even started.
                # Do not call the task run() method, just stop the thread.
                _logger.debug("Task thread %s stopped before start", self._task_name)
                return

        _logger.debug("Task thread %s starts running", self._task_name)

        try:
            # Invoke the task main function.
            self.task.run()
        except QMI_TaskStopException:
            # The task was stopped via QMI_TaskStopException.
            # Log this, but don't re-raise the exception.
            _logger.warning("Task %s stopped on QMI_TaskStopException", self._task_name, exc_info=True)
        except BaseException as exception:
            # The task main function raised an exception.
            # Report the exception and stop the thread.
            with self._state_cond:
                _logger.warning("Exception in task %s", self._task_name, exc_info=True)
                self._exception = exception
                self._state = _TaskThread.State.EXCEPTION_WHILE_RUNNING_TASK
                self._state_cond.notify_all()
                return

        # Notify the outside world that the task is finished.
        with self._state_cond:
            self._state = _TaskThread.State.TASK_COMPLETED_NORMALLY
            self._state_cond.notify_all()

        _logger.debug("Task thread %s completed normally", self._task_name)

    def get_state(self) -> Tuple[State, Optional[BaseException]]:
        """Return task state and any exception that occurred in the task."""
        with self._state_cond:
            return (self._state, self._exception)

    def wait_until_initialized(self) -> None:
        """Block until initialization of the task is finished."""
        with self._state_cond:
            while self._state == _TaskThread.State.INITIAL:
                self._state_cond.wait()

    def start_task(self) -> None:
        """Start the task.

        This causes invocation of the `QMI_Task.run()` method inside
        the dedicated thread.

        This function may be called only once per task.
        """
        with self._state_cond:
            # Wait until initialization finished.
            while self._state == _TaskThread.State.INITIAL:
                self._state_cond.wait()
            # Sanity check on task state.
            assert self._state == _TaskThread.State.READY_TO_RUN
            # Kick task to RUNNING state.
            self._state = _TaskThread.State.RUNNING
            self._state_cond.notify_all()

    def stop_task(self) -> None:
        """Tell the task to stop.

        This function signals the task to stop, then returns immediately.
        It may take some time until the task actually stops, depending on
        the code inside the task.

        If necessary, call `wait_until_complete()` to wait for actual
        completion of the task.
        """

        with self._state_cond:

            if self._state == _TaskThread.State.EXCEPTION_WHILE_INSTANTIATING_TASK:
                # Task already stopped.
                return

            if self._state in (_TaskThread.State.INITIAL, _TaskThread.State.READY_TO_RUN):
                # Task is still waiting to start. Stop it now.
                self._state = _TaskThread.State.TASK_STOPPED_BEFORE_START
                self._state_cond.notify_all()
                return

            # Sanity check on task state.
            assert self._state in (_TaskThread.State.RUNNING,
                                   _TaskThread.State.TASK_COMPLETED_NORMALLY,
                                   _TaskThread.State.EXCEPTION_WHILE_RUNNING_TASK,
                                   _TaskThread.State.TASK_STOPPED_BEFORE_START)
            assert self.task is not None

        # Set flag to make task code stop.
        self.task._stop_requested.set()

        # If the task is waiting on a condition, force it to wake up.
        with self._wait_cond_lock:
            # DO NOT hold self._wait_cond_lock while locking the condition lock.
            # Doing so could cause deadlock since] wait_for_condition() holds
            # the condition lock while locking self._wait_cond_lock.
            wait_cond = self._wait_cond

        if wait_cond is not None:
            with wait_cond:
                wait_cond.notify_all()

    def _request_shutdown(self) -> None:
        # Stop the task during Python shutdown.
        _logger.warning("Stopping task %s during shutdown", self._task_name)
        self.stop_task()

    def wait_for_condition(self,
                           cond: threading.Condition,
                           predicate: Callable[[], bool],
                           timeout: Optional[float]
                           ) -> bool:
        """Wait until a condition becomes true, or until the task is stopped.

        This method provides a way for QMI code to wait on a condition variable,
        but stop waiting if the running task receives a stop request.
        This is done by internally notifying the condition variable when a stop
        request is received.

        This function may only be called from the thread that is represented
        by this instance of `_TaskThread`.

        Parameters:
            cond:       Condition variable to wait on. The calling thread
                        must hold the lock associated with the condition
                        variable before calling this function.
            predicate:  Callable function which returns a boolean.
                        The predicate will be evaluated before waiting,
                        and evaluated again each time the condition variable
                        is notified. Waiting ends when the predicate
                        returns True.
            timeout:    Timeout in seconds, or None to wait indefinitely.

        Returns:
            True if the condition becomes true, False if timeout occurs.

        Raises:
            QMI_TaskStopException: If the thread receives a stop request before
                the condition becomes true.
        """

        # Check that this method is called from within this thread.
        assert threading.current_thread() is self
        assert self.task is not None

        # Register the condition variable on which we are going to wait.
        # This will be used to stop waiting if the task receives a stop request.
        with self._wait_cond_lock:
            assert self._wait_cond is None
            self._wait_cond = cond

        try:
            # Wait until either the predicate becomes true or the task receives a stop request.
            stop_requested = self.task._stop_requested
            ret = cond.wait_for(lambda: (predicate() or stop_requested.is_set()), timeout)

            # Raise QMI_TaskStopException if the task has received a stop request.
            if self.task._stop_requested.is_set():
                raise QMI_TaskStopException()

            # Otherwise return True if the predicate became True, False if waiting timed out.
            return ret

        finally:
            # Unregister the condition variable on which we waited.
            with self._wait_cond_lock:
                self._wait_cond = None


class QMI_TaskRunner(QMI_RpcObject):
    """Manager for a single task.

    A `QMI_TaskRunner` manages a single instance of `QMI_Task`.

    A dedicated `QMI_TaskRunner` is created by the context for each instance
    of `QMI_Task`. The `QMI_TaskRunner` creates a background thread to run
    the actual `QMI_Task` instance.

    The `QMI_TaskRunner` supports a fixed set of RPC methods.
    To interact with the task, other tasks or routines call these methods.

    A `QMI_TaskRunner` instance is created by the context when the
    application calls `context.make_task()`. Application code should
    not explicitly call the `QMI_TaskRunner()` constructor.
    """

    @classmethod
    def get_category(cls) -> Optional[str]:
        return "task"

    def __init__(self,
                 context: 'qmi.core.context.QMI_Context',
                 name: str,
                 task_class: Type[QMI_Task],
                 task_args: tuple,
                 task_kwargs: dict
                 ) -> None:
        """Initialize the task runner and create the task instance.

        If the task raises an exception during initialization,
        the task will be cleaned up and the exception re-raised from this method.

        Parameters:
            context: The QMI context.
            name: RPC object name of the task.
            task_class: Specific subclass of `QMI_Task` to be instantiated.
            task_args: Positional arguments for the `QMI_Task` constructor.
            task_kwargs: Keyword arguments for the `QMI_Task` constructor.
        """

        _logger.debug("Creating task %s", name)

        # An instance of QMI_TaskRunner publishes signals which are declared
        # in another class, namely in the specific subclass of QMI_Task
        # which is managed by the task runner. We therefore specify the
        # task class as the "signal_declaration_class".
        super().__init__(context, name, signal_declaration_class=task_class)

        self._task_class = task_class
        self._new_settings = None
        self._joined = False
        self._policy = None

        # Create thread and start it.
        self._thread = _TaskThread(self, name, task_class, task_args, task_kwargs)
        self._thread.start()

        # The thread will create an instance of the QMI_Task subclass.
        # Wait until the instance is initialized.
        self._thread.wait_until_initialized()
        (state, exception) = self._thread.get_state()

        if state == _TaskThread.State.EXCEPTION_WHILE_INSTANTIATING_TASK:
            # An exception occurred during initialization.
            # Clean up task and re-raise the exception here.
            self._thread.join()
            assert isinstance(exception, BaseException)
            raise QMI_TaskInitException("Failed to initialize task {}".format(self._name)) from exception

        assert state == _TaskThread.State.READY_TO_RUN
        assert self._thread.task is not None

    @rpc_method
    def __enter__(self):
        """The `__enter__` methods is decorated as `rpc_method` so that `QMI_RpcProxy` can call it when using the
        proxy with a `with` context manager. This method also calls to start the task thread."""
        return self.start()

    @rpc_method
    def __exit__(self, *args, **kwargs):
        """The `__exit__` methods is decorated as `rpc_method` so that `QMI_RpcProxy` can call it when using the
        proxy with a `with` context manager. This method also calls to stop and join the task thread."""
        self.stop()
        self.join()

    def release_rpc_object(self) -> None:
        """Ensure the task is joined before it is removed from the context."""
        if not self._joined:
            _logger.warning("Task %s removed but not joined; stopping it now", self._name)
            self.stop()
            self.join()

    @rpc_method
    def start(self) -> None:
        """Start the task.

        This function triggers a call to the `run()` method of the task.
        Note that the task runs in a separate thread.
        This function returns as soon as the task is started, while the
        task continues to run in the background.

        This function may be called only once per task.
        """

        # Check that task was not yet started.
        (state, dummy_exception) = self._thread.get_state()
        if state != _TaskThread.State.READY_TO_RUN:
            raise QMI_UsageException("Task {} can not be started more than once".format(self._name))

        _logger.debug("Starting task %s", self._name)
        self._thread.start_task()

    @rpc_method
    def stop(self) -> None:
        """Stop the task.

        This function tells the task to stop. From this point on, if the
        task calls `stop_requested()` it will return True and if the task
        sleeps it will be stopped via `QMI_TaskStopException`.

        This function returns immediately without waiting until the task
        is stopped. The task may continue to run in the background for
        some time, depending on the code inside the task.
        """
        _logger.debug("Stopping task %s", self._name)
        self._thread.stop_task()

    @rpc_method
    def join(self) -> None:
        """Wait until the task is fully stopped.

        This function blocks until the task is fully stopped.
        This may take some time, depending on the code inside the task.
        If the task does not stop spontaneously and `stop()` has not been
        called, this function may never return.

        If the task code raised an exception, the exception will
        be re-raised from this method.

        Each task must be joined before it can be removed from the context.
        Joining the task will also stop the associated thread and release its
        resources.
        """

        _logger.debug("Joining task %s", self._name)

        # Wait until the task thread ends.
        self._thread.join()

        # Get the task state.
        (state, exception) = self._thread.get_state()
        assert state in (_TaskThread.State.TASK_COMPLETED_NORMALLY,
                         _TaskThread.State.EXCEPTION_WHILE_RUNNING_TASK,
                         _TaskThread.State.TASK_STOPPED_BEFORE_START)

        # Break the reference cycle between QMI_TaskRunner and QMI_Task.
        assert self._thread.task is not None
        self._thread.task._task_runner = None  # type: ignore

        # Mark the task as joined (safe to be released).
        self._joined = True

        if state == _TaskThread.State.EXCEPTION_WHILE_RUNNING_TASK:
            # An exception occurred while running the task.
            # Re-raise the exception to report it to the application.
            assert isinstance(exception, BaseException)
            raise QMI_TaskRunException("Task {} failed".format(self._name)) from exception

    @rpc_method
    def is_running(self) -> bool:
        """Return True if the task is currently running.

        This method returns True if the task has been started, has not yet
        stopped and has not raised an exception.
        """
        (state, dummy_exception) = self._thread.get_state()
        return state == _TaskThread.State.RUNNING

    @rpc_method
    def set_settings(self, new_settings: Any) -> None:
        """Send new settings to the task.

        Other parts of the software may call this method to change the settings
        used by the task.

        Note that the task may not immediately use the new settings. The task
        continues to use its current settings until it explicitly updates the
        settings by calling `update_settings()`.
        """
        assert self._thread.task is not None
        self._thread.task._settings_fifo.append(new_settings)

    @rpc_method
    def get_settings(self) -> Any:
        """Return the current value of the task's `self.settings` variable.

        The task can receive settings from other parts of the software.
        This method returns a snapshot of the settings that are currently
        in effect.

        Note that calling `get_settings()` immediately after `set_settings()`
        may still return the old settings until the task explicitly updates
        its settings by calling `update_settings()`.
        """
        assert self._thread.task is not None
        return self._thread.task.settings

    @rpc_method
    def get_pending_settings(self) -> Any:
        """Return pending settings stored in the task runner's fifo.

        Only once the internal task calls `update_settings()`, is the pending
        settings from the fifo assigned to the task's actual settings.
        """
        assert self._thread.task is not None
        pending_settings = list(self._thread.task._settings_fifo)
        return pending_settings[0] if pending_settings else None

    @rpc_method
    def get_status(self) -> Any:
        """Return the current value of the task's `self.status` variable.

        The task code can update its `self.status` variable at any time
        to expose its current status to other parts of the software.
        This method returns a current snapshot of the status variable.
        """
        assert self._thread.task is not None
        return self._thread.task.status

    @rpc_method
    def get_task_class_name(self) -> str:
        """Return the name of the QMI_Task class managed by this task runner."""
        module_name = self._task_class.__module__
        class_name = self._task_class.__name__
        return module_name + "." + class_name


class QMI_LoopTask(QMI_Task):
    """
    QMI_LoopTask is a subclass of `QMI_Task`, implementing specific background loop task.
    It has its own special `run()` method, containing a sequence of actions that apply to most `QMI_Task`
    implementations that will be used.

    Attributes:
        sig_status_updated: Signal used to publish the latest status when updated. Can be used by clients who need
                              to know when the status might have changed by another client. Users need to re-define
                              this signal with appropriate type, if necessary.
    """
    sig_status_updated = QMI_Signal([type(None)])

    def __init__(self, task_runner, name, loop_period: float = 1.0,
                 policy: QMI_LoopTaskMissedLoopPolicy = QMI_LoopTaskMissedLoopPolicy.IMMEDIATE) -> None:
        """
        Initialize the loop task.

        A QMI_Task subclass, overriding the base class. The `__init__()` function takes `task_runner` and `name`,
        which are provided by QMI context. The subclass `__init__()` calls the base class method via
        ``super().__init__(task_runner, name)``.

        The subclass `__init__()` takes additional parameters `loop_period` and `policy`. These additional parameters
        can be passed to `context.make_task()`.

        Parameters:
            task_runner: QMI_TaskRunner - object instance which manages this task.
            name: string - The RPC object name of this task instance.
            loop_period: float - The loop task repeat period in (fractions of) seconds.
            policy: QMI_LoopTaskMissedLoopPolicy - The policy for handling missed loop periods
        """
        super().__init__(task_runner, name)
        self._loop_period = loop_period
        self._policy = policy
        self._status_fifo: collections.deque = collections.deque(maxlen=1)

        declared_signals = self._qmi_signals  # type: ignore  # pylint: disable=no-member
        # Create an instance of QMI_RegisteredSignal for sig_status_updated, published by this class.
        # Insert this QMI_RegisteredSignal instance as an attribute of the RpcObject instance.
        for sig_desc in declared_signals:
            if sig_desc.name == "sig_status_updated":
                # specifically remap settings because this is only typed in the inherited task
                sig = QMI_RegisteredSignal(self._context, name, sig_desc.name, tuple([type(self.status)]))
                setattr(self, sig_desc.name, sig)

    def run(self) -> None:
        """
        Main function inside the task thread. This overrides the base class implementation.

        The default implementation creates a skeleton of actions that a task loop will take. The clients using this
        QMI_LoopTask will have to override the implementations of `loop_prepare()`, `loop_iteration()`,
        `loop_finalize()`, `process_new_settings()`, `update_status()`, and `publish_signals()`.

        The behaviour of handling missed loop iterations depend on the `QMI_LoopTaskMissedLoopPolicy` attribute set.
        """

        _logger.info("[%s] Starting...", self._name)

        self.loop_prepare()
        next_time = time.monotonic() + self._loop_period
        try:
            while not self.stop_requested():

                # Check for updated settings. If updated, process them
                if self.update_settings():
                    self.process_new_settings()

                # Do work.
                self.loop_iteration()

                # Check for updates status. If updated, publish new status.
                if self.update_status():
                    self.sig_status_updated.publish(self.status)

                # Update other signals (if present).
                self.publish_signals()

                # Sleep until next iteration.
                time_to_sleep = next_time - time.monotonic()
                if time_to_sleep > 0:
                    self.sleep(time_to_sleep)
                    next_time += self._loop_period

                else:
                    _logger.warning("[%s] Missed loop time: %.3f seconds late", self._name, -time_to_sleep)
                    if self._policy == QMI_LoopTaskMissedLoopPolicy.IMMEDIATE:
                        # Try to do the next one as soon as possible, but do not miss any steps
                        next_time = time.monotonic() + self._loop_period

                    elif self._policy == QMI_LoopTaskMissedLoopPolicy.SKIP:
                        # Check how many loop periods were missed and set the next one after those
                        periods_missed = int((self._loop_period - time_to_sleep) / self._loop_period)
                        next_time += self._loop_period * periods_missed

                    elif self._policy == QMI_LoopTaskMissedLoopPolicy.TERMINATE:
                        # Stop loop
                        self._task_runner.stop()

        except QMI_TaskStopException:
            pass  # not an error

        finally:
            self.loop_finalize()

        _logger.info("[%s] Stopped", self._name)

    def loop_prepare(self) -> None:
        """Method to prepare the task loop.

        Subclasses may override this method to prepare the task loop.
        The default implementation does nothing.
        """
        pass

    def process_new_settings(self) -> None:
        """Method for processing new settings.

        Subclasses may override this method to new task settings.
        The default implementation does nothing.
        """
        pass

    def loop_iteration(self) -> None:
        """Define work to be done in the task loop.

        Subclasses may override this method to define work done in a loop iteration.
        The default implementation does nothing.
        """
        pass

    def update_status(self) -> bool:
        """Update to the latest status that were set to the task.

        If another routine called `set_status()` to set a new status to this task, this method should copy the new
        status to `self.status` and return True. Otherwise the status remains the same and this method returns False.

        Subclasses may override this method to set specific status.
        The default implementation does nothing (returns False).

        Returns:
             bool - True if there is a new status, False if the status is unchanged.
        """
        return False

    def publish_signals(self) -> None:
        """Method for publishing other signals described by the client for the loop task.

        Subclasses may override this method to publish specific signals.
        The default implementation does nothing.
        """
        pass

    def loop_finalize(self) -> None:
        """The loop should return the task to a specific state and settings at the end.

        Subclasses may override this method to specify finalizing actions.
        The default implementation does nothing.
        """
        pass


# Imports needed only for static typing.
if typing.TYPE_CHECKING:
    import qmi.core.context



"""QMI Instrument driver for the Tektronix's AWG 5014 arbitrary waveform generator."""
from pathlib import PureWindowsPath
import socket
from typing import List, Tuple, Union
import logging
import re
from time import time, sleep

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


AWG_FILE_BLOCK_DATA_LIMIT = 65E7


class Tektronix_Awg5014(QMI_Instrument):
    """QMI Instrument driver for the Tektronix AWG5014 Arbitrary Waveform Generator."""

    RESPONSE_TIMEOUT = 30
    NUM_CHANNELS = 4

    _instrument_status = {"0": "IDLE", "1": "WAIT_FOR_TRIGGER", "2": "RUNNING"}

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize the driver.

        To connect to the instrument via Ethernet, specify transport descriptor "tcp:<IP-address>:4000".
        To connect to the instrument using VXI11, specify transport descriptor "vxi11:<IP-address>".
        WARNING: The following transport options have not been tested but should in principle work.
        To connect to the instrument via GPIB, specify transport descriptor "GPIB0:1:INSTR".
        To connect to the instrument via USB, specify transport descriptor
          "usbtmc:vendorid=0x0699:productid=0x0356:serialnr=<serial_number>".
        :END WARNING

        Parameters:
            context: The QMI context the instrument proxy is located.
            name: Name for this instrument instance.
            transport: Transport descriptor to access this instrument.
        """
        super().__init__(context, name)

        self._transport = create_transport(transport)
        self._scpi_transport = ScpiProtocol(self._transport, default_timeout=self.RESPONSE_TIMEOUT)

    def _check_error(self) -> None:
        """Read the instrument error queue and raise an exception if there is an error."""
        # Read error queue.
        resp = self._scpi_transport.ask("SYST:ERR?", discard=True)
        if not re.match(r"^\s*[+-]?0\s*,", resp):
            # Report the error.
            raise QMI_InstrumentException("Instrument returned error: {}".format(resp))

    def _check_channel_number(self, channel: int) -> None:
        if channel not in range(1, self.NUM_CHANNELS + 1):
            raise ValueError("Channel number must be in range [1-4].")

    def _get_memory_catalog(self) -> str:
        """Get a string with the total amount of storage used [bytes], free space left in the mass storage [bytes], and
        a list of file names and directories from instrument (from current directory).

        Returns:
            string <storage used>,<storage free>[,<file name>,<file type>,<file size>]*n for n files.
        """
        return self._scpi_transport.ask("MMEM:CAT?")

    @rpc_method
    def open(self) -> None:
        """Open the connection to the instrument."""
        self._check_is_closed()
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        """Close the connection to the instrument."""
        self._check_is_open()
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def wait_and_clear(self) -> None:
        """Wait and clear (responses of) previous command(s)."""
        # Wait until all pending commands are finished.
        self._scpi_transport.write("*WAI")
        try:
            # Clear out any possible hanging responses from the Tektronix
            self._scpi_transport.write("*CLS")

        except (QMI_TimeoutException, socket.timeout):
            # In case no hanging responses were present
            pass

        finally:
            # Clear transport buffer
            self._transport.discard_read()

    @rpc_method
    def wait_command_completion(self) -> None:
        """Wait completion of previous command."""
        self._scpi_transport.write("*OPC")
        start = time()
        try:
            while (time() - start) < self.RESPONSE_TIMEOUT:
                operation_complete = self._scpi_transport.ask("*OPC?")
                if operation_complete == "1":
                    break

                sleep(1)

        except (QMI_TimeoutException, socket.timeout):
            # We ignore timeout errors and go to check if there were errors from the controller
            pass

        finally:
            # Clear the error queue from "error" -800,"Operation complete"
            resp = self._scpi_transport.ask("SYST:ERR?", discard=True)
            if not re.match(r"^\s*-800\s*,", resp) and not re.match(r"^\s*[+-]?0\s*,", resp):
                # Report the error.
                raise QMI_InstrumentException(f"Instrument returned error: {resp}")

    @rpc_method
    def reset(self) -> None:
        """Resets the instrument to default values. In practise, it was seen that *CLS does not clear the memory as
        intended for stale responses. By making this as a query with discarding data in transport buffer, it clears
        up possible delayed response(s) from the scpi_protocol.
        """
        _logger.info("Resetting instrument [%s]", self._name)
        # Do reset
        self._scpi_transport.write("*RST")
        # Re-enable OPC
        self._scpi_transport.write("*OPC")
        try:
            self._scpi_transport.write("*CLS")
            self._transport.discard_read()

        except (QMI_TimeoutException, socket.timeout):
            # In case no hanging responses were present
            pass

    @rpc_method
    def get_error(self) -> None:
        """Allow client to poll for errors. Any errors will be raised as exceptions.

        Raises:
            QMI_InstrumentException with any error code other than 0.
        """
        self._check_error()

    @rpc_method
    def get_state(self) -> str:
        """Queries the instrument state. Possible responses are 0: IDLE, 1: WAIT_FOR_TRIGGER, 2: RUNNING.

        Returns:
            Instrument state translated from dictionary.
        """
        state = self._scpi_transport.ask("AWGC:RSTATE?")
        return self._instrument_status[state.strip()]

    @rpc_method
    def get_setup_file_name(self) -> Tuple[str, str]:
        """Returns the current setup file name of the arbitrary waveform generator.
           The response contains the full path for the file including the disk drive.
           Example: '"\\my\\project\\awg\\setup\\a1.awg","D:"'

        Returns:
            tuple: File path and name, e.g. ("\\Users\\OEM\\Documents", "newAWGfile.awg").
        """
        resp = PureWindowsPath(self._scpi_transport.ask("AWGC:SNAM?"))
        return str(resp.parent).lstrip('"'), resp.name.split(",")[0].rstrip('"')

    @rpc_method
    def get_file_names(self) -> List[str]:
        """Get a string with a list of file names from instrument.

        Returns:
            List [<file name>,<file type>,<file size>]*n for n files.
        """
        catalog = self._get_memory_catalog()
        catalog_list = catalog.split('"')[1:]  # drop off the storage info
        files = ""
        for entry in catalog_list:
            if ",DIR," not in entry and len(entry) > 1:
                files = files + f"{entry};"

        return files.rstrip(";").split(";")  # Remove last semi-colon and split into a list

    @rpc_method
    def get_directories(self) -> List[str]:
        """Return the directory contents from memory.

        Returns:
            List [<dir name>]*n for n directories.
        """
        catalog = self._get_memory_catalog()
        catalog_list = catalog.split('"')[1:]  # drop off the storage info
        directories = ""
        for entry in catalog_list:
            if ",DIR," in entry:
                dir_name = entry.split(",")[0]
                directories = directories + f"{dir_name};"

        return directories.rstrip(";").split(";")  # Remove last semi-colon

    @rpc_method
    def get_current_directory_name(self) -> str:
        """Return the current directory name from memory.

        Returns:
            Current directory name.
        """
        return self._scpi_transport.ask("MMEM:CDIR?").strip()

    @rpc_method
    def change_directory(self, cd: str) -> None:
        """Change the current directory.

        Parameters:
            cd: The new directory to cd to, in form "\\DIR1\\DIR2" or "C:\\DIR1\\DIR2" or ".." etc.
        """
        self._scpi_transport.write(f'MMEM:CDIR "{cd}"')

    @rpc_method
    def cd_to_root(self) -> None:
        """Change directory to root."""
        self._scpi_transport.write('MMEM:CDIR "C:\\.."')

    @rpc_method
    def create_directory(self, new_dir: str) -> None:
        """Create a new subdirectory in the current directory.

        Parameters:
            new_dir: The new directory name
        """
        # First check that we don't try to create a duplicate
        directories = self.get_directories()
        if new_dir in directories:
            raise QMI_InstrumentException("Error: Directory already exists.")

        self._scpi_transport.write(f'MMEM:MDIR "{new_dir}"')

    @rpc_method
    def remove_directory(self, del_dir: str) -> None:
        """Remove a subdirectory in the current directory.

        Parameters:
            del_dir: The deletable directory name
        """
        # First check that we have a directory named as the deletable one
        directories = self.get_directories()
        if del_dir in directories:
            self._scpi_transport.write(f'MMEM:DEL "{del_dir}"')
            return

        raise QMI_InstrumentException("Error: Directory does not exist.")

    @rpc_method
    def send_awg_file(self, file_name: str, awg_file_block_data: bytes) -> None:
        """This command creates a file on the AWG from the input block data. The file is created with the given name
        into the current directory the AWG files system is on. The waveform data should be in integers (-1, 0, 1) as
        only this format is currently supported. Floats could be needed when one needs to have correct waveform when
        waveform contains more than (-1, 0, 1) (+/-peak and 0) values. Using integer format is faster. Note that the
        "MMEM[ory]:DATA has a limit of 650 000 000 bytes of data.

        From the "AWG5000 and AWG7000 Series Arbitrary Waveform Generators Programmer Manual", table 2-4, data block is
        explained as string:
            #512234xxxxx... where 5 indicates that the following 5 digits (12234) specify the length of the data in
            bytes; xxxxx... indicates actual data or #0xxxxx...<LF><&EOI>.

        As this command is slow with large AWG files, successive command calls without waiting for the sending to
        complete can cause errors. Therefore, it is recommended to call `wait_command_completion` after this command.

        Parameters:
            file_name:           The file name, e.g. "newAWGfile.awg".
            awg_file_block_data: Block data for the file in bytes, should be sum of values of io.BytesIO objects
                headers + channel records + waveform records + sequence records. Currently only integer format.
        """
        data_length: int = len(awg_file_block_data)
        if data_length > AWG_FILE_BLOCK_DATA_LIMIT:
            raise QMI_InstrumentException("Waveform data block length too large! Consider using more efficient " +
                                          "file encoding or direct control commands on instrument.")

        block_data = f"#{len(str(data_length))}{data_length}".encode("ascii") + awg_file_block_data
        self._scpi_transport.write_raw(f'MMEM:DATA "{file_name}",'.encode("ascii") + block_data)

    @rpc_method
    def load_awg_file(self, file_name: str) -> None:
        """This command shall load the given .awg file into the AWG memory. Note that it will also overwrite
        all the current instrument settings with the values in the file.

        As this command is slow with large AWG files, successive command calls without waiting for the upload to
        complete can cause errors. Therefore, it is recommended to call `wait_command_completion` after this command.

        Parameters:
            file_name: The file name, e.g. "newAWGfile.awg".
        """
        # First check that we have a file named as the input file
        files = self.get_file_names()
        for f in files:
            if file_name in f:
                self._scpi_transport.write(f'AWGC:SRES "{file_name}"')
                return

        raise QMI_InstrumentException(f"Error: File {file_name} not found.")

    @rpc_method
    def remove_file(self, del_file: str) -> None:
        """Remove a file in the current directory.

        Parameters:
            del_file: The deletable file name
        """
        # First check that we have a file named as the deletable one
        files = self.get_file_names()
        for f in files:
            if del_file in f:
                self._scpi_transport.write(f'MMEM:DEL "{del_file}"')
                return

        raise QMI_InstrumentException("Error: File does not exist.")

    @rpc_method
    def clear_waveforms(self) -> None:
        """Clears the waveforms on all channels by settings the waveform files to be '' (empty string)."""
        for i in range(1, self.NUM_CHANNELS + 1):
            self._scpi_transport.write(f'SOUR{i}:FUNC:USER ""')

    @rpc_method
    def delete_all_waveforms_from_list(self) -> None:
        """Deletes all waveforms from memory list."""
        self._scpi_transport.write("WLIS:WAV:DEL ALL")

    @rpc_method
    def start(self) -> None:
        """Initiates the output of a waveform or a sequence.

        As this command can run for a relatively long time, successive command calls can cause errors. Therefore,
        it is recommended to call `wait_command_completion` after this command.
        """
        self._scpi_transport.write("AWGC:RUN")

    @rpc_method
    def stop(self) -> None:
        """Stops the output of a waveform or a sequence."""
        self._scpi_transport.write("AWGC:STOP")

    @rpc_method
    def get_clock_source(self) -> str:
        """Query the AWG whether the 10MHz is set to the internal or external source.

        Returns:
            "INT" or "EXT"
        """
        return self._scpi_transport.ask("AWGC:CLOC:SOUR?")

    @rpc_method
    def set_clock_to_internal(self) -> None:
        """Set the clock source to be internal."""
        self._scpi_transport.write(f"AWGC:CLOC:SOUR INT")

    @rpc_method
    def set_clock_to_external(self) -> None:
        """Set the clock source to be external."""
        self._scpi_transport.write(f"AWGC:CLOC:SOUR EXT")

    @rpc_method
    def get_reference_oscillator_source(self) -> str:
        """Query the AWG whether the 10MHz reference is set to the internal or external source.

        Returns:
            "INT" or "EXT"
        """
        return self._scpi_transport.ask("ROSC:SOUR?").strip()

    @rpc_method
    def set_reference_oscillator_to_internal(self) -> None:
        """Set the 10MHz reference source to be internal."""
        self._scpi_transport.write(f"ROSC:SOUR INT")

    @rpc_method
    def set_reference_oscillator_to_external(self) -> None:
        """Set the 10MHz reference source to be external."""
        self._scpi_transport.write(f"ROSC:SOUR EXT")

    @rpc_method
    def get_reference_oscillator_type(self) -> str:
        """Query the type of the reference oscillator.

        Returns:
            Reference oscillator type.
        """
        return self._scpi_transport.ask("ROSC:TYPE?").strip()

    @rpc_method
    def set_reference_oscillator_type(self, type: str) -> None:
        """Set the reference oscillator type to fixed or variable.

        Parameters:
              Reference oscillator type ["FIX" or "VAR"]
        """
        if type.upper() not in ["FIX", "VAR"]:
            raise ValueError(f"Unknown reference oscillator type {type}!")

        self._scpi_transport.write(f"ROSC:TYPE {type}")

    @rpc_method
    def get_reference_clock_frequency(self) -> float:
        """Query the reference clock frequency.

        Returns:
            The reference clock frequency in Hertz. Possible values 1E7, 2E7, 1E8.
        """
        return float(self._scpi_transport.ask("ROSC:FREQ?"))

    @rpc_method
    def set_reference_clock_frequency(self, frequency: float) -> None:
        """Set the reference clock frequency. The possible values are 10, 20 and 100 MHz.

        Parameters:
            frequency: The new reference clock frequency in Hertz.
        """
        mhz_dict = {1E7: "10MHZ", 2E7: "20MHZ", 10E7: "100MHZ"}
        if frequency not in mhz_dict.keys():
            raise ValueError("Error: New reference clock frequency is any of valid values [10, 20, 100] MHz!")

        self._scpi_transport.write(f"ROSC:FREQ {mhz_dict[frequency]}")

    @rpc_method
    def get_source_clock_frequency(self) -> float:
        """Query the source clock frequency.

        Returns:
            The source clock frequency in Hertz.
        """
        return float(self._scpi_transport.ask("SOUR:FREQ?"))

    @rpc_method
    def set_source_clock_frequency(self, frequency: float) -> None:
        """Set the source clock frequency. This will work either if "INT" is selected as reference source or
        reference source is set as "EXT" and reference type as "FIX".

        Parameters:
            frequency: The new source clock frequency in Hertz.
        """
        if frequency < 10E6 or frequency > 10E9:
            raise ValueError("Error: New source clock frequency is out of valid range of 10Mhz and 10GHz!")

        if self.get_reference_oscillator_source() == "INT":
            self._scpi_transport.write(f"SOUR:FREQ {frequency:.7G}")

        elif self.get_reference_oscillator_type() == "FIX":
            self._scpi_transport.write(f"SOUR:FREQ:FIX {frequency:.7G}")

        else:
            raise RuntimeError("Error: Can set frequency either when source oscillator source is 'INT', or " +
                               "when source oscillator is 'EXT' and source oscillator type is 'FIX'.")

    @rpc_method
    def get_amplitude(self, channel: int) -> float:
        """Get the signal amplitude of a waveform associated with specific channel.

        Parameters:
            channel: The channel number [1-4].

        Returns:
            The peak-to-peak signal amplitude in Volts.
        """
        self._check_channel_number(channel)
        return float(self._scpi_transport.ask(f"SOUR{channel}:VOLT:LEV:IMM:AMPL?"))

    @rpc_method
    def set_amplitude(self, channel: int, amplitude: float) -> None:
        """Set the signal amplitude for a waveform associated with specific channel.

        Parameters:
            channel: The channel number [1-4].
            amplitude: The target amplitude voltage (peak-to-peak).
        """
        self._check_channel_number(channel)
        self._scpi_transport.write(f"SOUR{channel}:VOLT:LEV:IMM:AMPL {amplitude:.6f}")

    @rpc_method
    def get_offset(self, channel: int) -> float:
        """Get the offset of a specific source channel.

        Parameters:
            channel: The channel number [1-4].

        Returns:
            The offset in Volts.
        """
        self._check_channel_number(channel)
        return float(self._scpi_transport.ask(f"SOUR{channel}:VOLT:LEV:IMM:OFFS?"))

    @rpc_method
    def set_offset(self, channel: int, offset: float) -> None:
        """Set the offset of a specific source channel.

        Parameters:
            channel: The channel number [1-4].
            offset: The offset in Volts
        """
        self._check_channel_number(channel)
        self._scpi_transport.write(f"SOUR{channel}:VOLT:LEV:IMM:OFFS {offset:.6f}")

    @rpc_method
    def get_dc_output_state(self) -> int:
        """Query if DC output state is 'on' or 'off'. The state is equal to all four DC outputs.

        Returns:
            DC state with 0 as 'off' and 1 as 'on'.
        """
        return int(self._scpi_transport.ask("AWGC:DC:STAT?"))

    @rpc_method
    def set_dc_output_state(self, state: int) -> None:
        """Set DC state as 'on' or 'off'. The command sets the same state for all four DC outputs.

        Parameters:
            state: DC state with 0 as 'off' and 1 as 'on'.
        """
        self._scpi_transport.write(f"AWGC:DC:STAT {state}")

    @rpc_method
    def get_dc_output_offset(self, output: int) -> float:
        """Get a DC output voltage offset.

        Parameters:
            output: DC output number in range [1-4].

        Returns:
            The offset in Volts.
        """
        self._check_channel_number(output)
        return float(self._scpi_transport.ask(f"AWGC:DC{output}:VOLT:OFFS?"))

    @rpc_method
    def set_dc_output_offset(self, output: int, offset: float) -> None:
        """Set a DC output voltage offset.

        Parameters:
            output: DC output number in range [1-4].
            offset: The offset in Volts, should be between [-3.0V-+5.0V].
        """
        self._check_channel_number(output)
        self._scpi_transport.write(f"AWGC:DC{output}:VOLT:OFFS {offset}V")

    @rpc_method
    def get_raw_dac_waveform_output(self, channel: int) -> int:
        """Get raw DAC waveform output channel state.

        Parameters:
            channel: DAC output channel number in range [1-4].

        Returns:
            Channel state with 0 as 'disabled' and 1 as 'enabled'.
        """
        self._check_channel_number(channel)
        return int(self._scpi_transport.ask(f"AWGC:DOUT{channel}:STAT?"))

    @rpc_method
    def set_raw_dac_waveform_output(self, channel: int, state: int) -> None:
        """Set raw DAC waveform output channel state. Channel state with 0 as 'disabled' and 1 as 'enabled'.

        Parameters:
            channel: DAC output channel number in range [1-4].
            state: Channel state with 0 as 'disabled' and 1 as 'enabled'.
        """
        self._check_channel_number(channel)
        self._scpi_transport.write(f"AWGC:DOUT{channel}:STAT {state}")

    @rpc_method
    def get_run_mode(self) -> str:
        """Get the current run mode.

        Returns:
            mode: The run mode.
        """
        return self._scpi_transport.ask("AWGC:RMOD?").strip()

    @rpc_method
    def set_run_mode_to_triggered(self) -> None:
        """Set the run mode to TRIGgered."""
        self._scpi_transport.write("AWGC:RMOD TRIG")

    @rpc_method
    def set_run_mode_to_continuous(self) -> None:
        """Set the run mode to CONTinuous."""
        self._scpi_transport.write("AWGC:RMOD CONT")

    @rpc_method
    def set_run_mode(self, run_mode: str) -> None:
        """Set new run mode.

        Parameters:
            run_mode: new run mode. Must be one of 'CONT', 'TRIG', 'SEQ' or 'GAT'. Also 'ENH' is possible for
                      compatibility reasons with other model series. 'ENH' is equal to 'SEQ'.
        """
        if run_mode.upper() not in ["CONT", "TRIG", "SEQ", "GAT", "ENH"]:
            raise ValueError(f"Invalid run mode {run_mode}")

        self._scpi_transport.write(f"AWGC:RMOD {run_mode}")

    @rpc_method
    def get_signal_addition(self) -> str:
        """Query the signal addition from external input state.

        Returns:
            The signal addition state. "ESIG" for enabled, "" for disabled.
        """
        return self._scpi_transport.ask("COMB:FEED?").strip()

    @rpc_method
    def set_signal_addition(self, state: str) -> None:
        """Set the signal addition from external input state.

        Parameters:
            state: The signal addition state. "ESIG" for enabled, "" for disabled.
        """
        self._scpi_transport.write(f'COMB:FEED "{state}"')

    @rpc_method
    def get_trigger_source(self) -> str:
        """Query the current trigger source."""
        return self._scpi_transport.ask("TRIG:SOUR?").strip()

    @rpc_method
    def set_trigger_source_to_internal(self) -> None:
        """Set the trigger source to INTernal."""
        self._scpi_transport.write("TRIG:SOUR INT")

    @rpc_method
    def set_trigger_source_to_external(self) -> None:
        """Set the trigger source to EXTernal."""
        self._scpi_transport.write("TRIG:SOUR EXT")

    @rpc_method
    def get_trigger_impedance(self) -> int:
        """Get the current external trigger impedance in Ohm.

        Returns:
            Trigger impedance value.
        """
        return int(float(self._scpi_transport.ask("TRIG:IMP?")))

    @rpc_method
    def set_trigger_impedance(self, impedance: int = 1000) -> None:
        """Sets the EXTERNAL trigger impedance to given value. Default value is 1000 Ohm. 50 Ohm is for the external
        trigger input.

        Parameters:
            impedance: The trigger impedance in Ohm. Possible values are 50 and 1000.
        """
        if impedance not in [50, 1000]:
            raise ValueError(f"Invalid impedance {impedance} Ohm. Must be either 50 Ohm or 1000 Ohm.")

        self._scpi_transport.write(f"TRIG:IMP {impedance}")

    @rpc_method
    def get_trigger_level(self) -> float:
        """Get current trigger level in Volts.

        Returns:
            Current trigger level.
        """
        return float(self._scpi_transport.ask("TRIG:LEV?"))

    @rpc_method
    def set_trigger_level(self, trigger_level: float) -> None:
        """Set a new trigger level. The value is rounded to 3 significant decimals.

        Parameters:
            trigger_level: The new trigger level in Volts.
        """
        self._scpi_transport.write(f"TRIG:LEV {trigger_level:.3f}")

    @rpc_method
    def get_trigger_slope(self) -> str:
        """Get trigger slope sign.

        Returns:
            Trigger slope sign as "POS" or "NEG".
        """
        return self._scpi_transport.ask("TRIG:SLOP?").strip()

    @rpc_method
    def set_trigger_slope(self, slope: str) -> None:
        """Set trigger slope sign.

        Parameters:
            Trigger slope sign as string "POS" or "POSitive" for positive slope and "NEG" or "NEGative" for
            negative slope.
        """
        if slope.upper() in ["POS", "POSITIVE"]:
            target_slope = "POSitive"

        elif slope.upper() in ["NEG", "NEGATIVE"]:
            target_slope = "NEGative"

        else:
            raise ValueError(f"Invalid slope sign {slope}")

        self._scpi_transport.write(f"TRIG:SLOP {target_slope}")

    @rpc_method
    def get_trigger_polarity(self) -> str:
        """Get trigger polarity sign.

        Returns:
            Trigger polarity sign as "POS" or "NEG".
        """
        return self._scpi_transport.ask("TRIG:POL?").strip()

    @rpc_method
    def set_trigger_polarity(self, polarity: str) -> None:
        """Set trigger polarity sign.

        Parameters:
            polarity: Trigger polarity sign as string "POS" or "POSitive" for positive polarity and "NEG" or "NEGative"
                      for negative polarity.
        """
        if polarity.upper() in ["POS", "POSITIVE"]:
            target_polarity = "POSitive"

        elif polarity.upper() in ["NEG", "NEGATIVE"]:
            target_polarity = "NEGative"

        else:
            raise ValueError(f"Invalid polarity sign {polarity}")

        self._scpi_transport.write(f"TRIG:POL {target_polarity}")

    @rpc_method
    def get_waveform_output_data_position(self) -> str:
        """Query The output data position of a waveform while the instrument is in the waiting-for-trigger state.
        This is valid only when Run Mode is Triggered or Gated.

        Returns:
            Waveform output data position as "FIRS" or "LAST".
        """
        return self._scpi_transport.ask("TRIG:SEQ:WVAL?").strip()

    @rpc_method
    def set_waveform_output_data_position(self, position: str) -> None:
        """
        This command sets the output data position of a waveform. This is valid only when
        Run Mode is Triggered or Gated.

        Parameters:
            position: Sets a position of a waveform as the output level. Possible values are "FIRS[t]" and "LAST".
        """
        if position.upper() not in ["FIRS", "FIRST", "LAST"]:
            raise ValueError(f"Invalid waveform position {position}")

        # Check the current Run mode before allowing to set
        if self.get_run_mode() not in ["TRIG", "GAT"]:
            raise QMI_InstrumentException("The Run mode must be Triggered or Gated to set WF output data position")

        self._scpi_transport.write(f"TRIG:SEQ:WVAL {position}")

    @rpc_method
    def force_trigger_event(self) -> None:
        """Generate a trigger event."""
        self._scpi_transport.write("*TRG")

    @rpc_method
    def get_channel_state(self, channel: int) -> bool:
        """Query if a channel is 'on' or 'off.

        Parameters:
            channel: Channel number in range [1-4].

        Returns:
            The channel state with False if 'off' and True if 'on'.
        """
        self._check_channel_number(channel)
        return bool(int(self._scpi_transport.ask(f"OUTP{channel}?")))

    @rpc_method
    def set_channel_state(self, channel: int, state: bool) -> None:
        """Set a channel state 'on' or 'off.

        Parameters:
            channel: Channel number in range [1-4].
            state: The new channel state with False if 'off' and True if 'on'.
        """
        set_state = "ON" if state else "OFF"
        self._check_channel_number(channel)
        self._scpi_transport.write(f"OUTP{channel} {set_state}")

    @rpc_method
    def get_low_pass_filter_frequency(self, channel: int) -> float:
        """Query low-pass filter frequency of an output channel.

        Parameters:
            channel: Channel number in range [1-4].

        Returns:
            The output channel filter frequency in MHz (9.9e37 at INFinity).
        """
        self._check_channel_number(channel)
        return float(self._scpi_transport.ask(f"OUTP{channel}:FILT:FREQ?"))

    @rpc_method
    def set_low_pass_filter_frequency(self, channel: int, frequency: Union[float, str]) -> None:
        """Set a low-pass filter frequency of an output channel frequency.

        Parameters:
            channel: Channel number in range [1-4].
            frequency: The new channel frequency in Hertz. Or "INF".
        """
        self._check_channel_number(channel)
        if isinstance(frequency, float):
            self._scpi_transport.write(
                f"OUTP{channel}:FILT:FREQ {frequency:.1E}".replace("E+0", "E+").replace("E-0", "E0")
            )

        elif isinstance(frequency, str):
            if frequency.upper() not in ["INF", "INFINITY"]:
                raise ValueError(f"Low pass filer frequency cannot be {frequency}, use a float or 'INFinity'")

            self._scpi_transport.write(f"OUTP{channel}:FILT:FREQ INFinity")

    @rpc_method
    def get_sequence_length(self) -> int:
        """Get sequence length.

        Returns:
            The sequence length.
        """
        return int(self._scpi_transport.ask("SEQ:LENG?"))

    @rpc_method
    def set_sequence_length(self, length: int) -> None:
        """Set sequence length. Note: length 0 will delete the sequence.

        Parameters:
            length: The sequence length.
        """
        self._scpi_transport.write(f"SEQ:LENG {length}")

    @rpc_method
    def get_sequencer_type(self) -> str:
        """Get the sequencer type to see if instrument is in hardware or software sequencer mode.

        Returns:
            Sequencer type, either "HARD" or "SOFT".
        """
        return self._scpi_transport.ask("AWGC:SEQ:TYPE?").strip()

    @rpc_method
    def get_sequencer_position(self) -> int:
        """Get the sequencer position.

        Returns:
            Sequencer position.
        """
        return int(self._scpi_transport.ask("AWGC:SEQ:POS?"))

    @rpc_method
    def set_sequence_element_goto_target_index(self, element_no: int, index_no: int) -> None:
        """Set target 'GOTO' index of a sequencer element. Note: First element is 1:

        Parameters:
            element_no: Element number.
            index_no: Target goto index number.
        """
        if index_no == 0:
            raise ValueError("Error: The sequence element indexing starts from 1, not from 0!")

        self._scpi_transport.write(f"SEQ:ELEM{element_no}:GOTO:IND {index_no}")

    @rpc_method
    def set_sequence_element_goto_state(self, element_no: int, goto_state: int) -> None:
        """Set 'GOTO' state for a sequencer element.

        Parameters:
            element_no: Element number.
            goto_state: New 'GOTO' state. 0 is 'off' and 1 is 'on'.
        """
        self._scpi_transport.write(f"SEQ:ELEM{element_no}:GOTO:STAT {goto_state}")

    @rpc_method
    def get_sequence_element_loop_count_infinite_state(self, element_no: int) -> bool:
        """Query the loop count infinity state of a sequencer element.

        Parameters:
            element_no: Element number.

        Returns:
            boolean, if the loop count to inf is 'on' = True or 'off' = False.
        """
        resp = self._scpi_transport.ask(f"SEQ:ELEM{element_no}:LOOP:INF?")
        return bool(int(resp))

    @rpc_method
    def set_sequence_element_loop_count_infinite_state(self, element_no: int, on: bool = True) -> None:
        """Set the loop count to infinity state for a sequencer element. Can be 'on' or 'off'.

        Parameters:
            element_no: Element number.
            on: boolean to set loop count to inf 'on' (True) or 'off' (False).
        """
        if on:
            self._scpi_transport.write(f"SEQ:ELEM{element_no}:LOOP:INF 1")

        else:
            self._scpi_transport.write(f"SEQ:ELEM{element_no}:LOOP:INF 0")

    @rpc_method
    def get_sequence_element_loop_count(self, element_no: int) -> int:
        """Get the loop count from a sequencer element.

        Parameters:
            element_no: Element number.

        Returns:
            Sequence element loop count.
        """
        return int(self._scpi_transport.ask(f"SEQ:ELEM{element_no}:LOOP:COUN?"))

    @rpc_method
    def set_sequence_element_loop_count(self, element_no: int, loop_count: int) -> None:
        """Set the loop count to given value for a sequencer element.

        Parameters:
            element_no: Element number.
            loop_count: The new loop count value in the range of [1, 65536]
        """
        if loop_count < 1 or loop_count > 65536:
            raise ValueError(f"Error: Invalid loop count value {loop_count}.")

        self._scpi_transport.write(f"SEQ:ELEM{element_no}:LOOP:COUN {loop_count}")

    @rpc_method
    def get_sequence_element_trigger_wait_state(self, element_no: int) -> int:
        """Get trigger wait state from a sequence element.

        Parameters:
            element_no: Element number.

        Returns:
            The sequence element trigger wait state. 0 for 'off', 1 for 'on'.
        """
        return int(self._scpi_transport.ask(f"SEQ:ELEM{element_no}:TWA?"))

    @rpc_method
    def set_sequence_element_trigger_wait_state(self, element_no: int, state: int) -> None:
        """Set trigger wait value state for a sequence element.

        Parameters:
            element_no: Element number.
            state: Trigger wait state. 0 for 'off', 1 for 'on'.
        """
        self._scpi_transport.write(f"SEQ:ELEM{element_no}:TWA {state}")

    @rpc_method
    def get_sequence_element_waveform(self, element_no: int, channel: int) -> str:
        """Get waveform from a sequence element channel.

        Parameters:
            element_no: Element number.
            channel: Channel number.

        Returns:
            The waveform type of the sequence element channel.
        """
        self._check_channel_number(channel)
        return self._scpi_transport.ask(f"SEQ:ELEM{element_no}:WAV{channel}?", discard=True).strip()

    @rpc_method
    def set_sequence_element_waveform(self, element_no: int, channel: int, waveform: str) -> None:
        """Set waveform for a sequence element channel.

        Parameters:
            element_no: Element number.
            channel: Channel number.
            waveform: Waveform type.
        """
        self._check_channel_number(channel)
        self._scpi_transport.write(f'SEQ:ELEM{element_no}:WAV{channel} "{waveform}"')

    @rpc_method
    def force_sequence_jump_to_index(self, jump_index_no: int) -> None:
        """Force sequence to jump to given index immediately.

        Parameters:
            jump_index_no: Target index number to jump to.
        """
        self._scpi_transport.write(f"SEQ:JUMP:IMM {jump_index_no}")

    @rpc_method
    def set_sequence_element_jump_target_index(self, element_no: int, jump_target_index: int) -> None:
        """Set sequence element event jump target index.

        Parameters:
            element_no: Element number.
            jump_target_index: The target index number.
        """
        self._scpi_transport.write(f"SEQ:ELEM{element_no}:JTAR:INDEX {jump_target_index}")

    @rpc_method
    def get_sequence_element_jump_type(self, element_no: int) -> str:
        """Get sequence element event jump type.

        Parameters:
            element_no: Element number.

        Returns:
            The jump type.
        """
        return self._scpi_transport.ask(f"SEQ:ELEM{element_no}:JTAR:TYPE?").strip()

    @rpc_method
    def set_sequence_element_jump_type(self, element_no: int, jump_type: str) -> None:
        """Set sequence element event jump type.

        Parameters:
            element_no: Element number.
            jump_type: The new jump type. Valid values are 'IND', 'NEXT', and 'OFF'.
        """
        if jump_type.upper() not in ["IND", "NEXT", "OFF"]:
            raise ValueError("Error: Invalid event jump target type.")

        self._scpi_transport.write(f"SEQ:ELEM{element_no}:JTAR:TYPE {jump_type}")

    @rpc_method
    def get_sequence_jump_mode(self) -> str:
        """Queries the sequence jump mode.

        Returns:
            Jump mode.
        """
        return self._scpi_transport.ask("AWGC:ENH:SEQ:JMOD?").strip()

    @rpc_method
    def set_sequence_jump_mode(self, jump_mode: str) -> None:
        """Set the sequence jump mode.

        Parameters:
            jump_mode: The jump mode, valid values are "LOG", "TABL" and "SOFT".
        """
        if jump_mode not in ["LOG", "TABL", "SOFT"]:
            raise ValueError(f"Error: Invalid jump mode {jump_mode}.")

        self._scpi_transport.write(f"AWGC:ENH:SEQ:JMOD {jump_mode}")

    @rpc_method
    def get_jump_target_definition(self, pattern: int) -> int:
        """Queries the dynamic jump target associated to specified event pattern.

        Parameters:
            pattern: The event pattern in range [0-511].

        Returns:
            The sequence index number associated to the event pattern.
        """
        return int(self._scpi_transport.ask(f"AWGC:EVEN:DJUM:DEF? {pattern}"))

    @rpc_method
    def set_jump_target_definition(self, pattern: int, jump_target: int) -> str:
        """Queries the dynamic jump target associated to specified event pattern.

        Parameters:
            pattern: The event pattern in range [0-511].
            jump_target: The target sequence index value, allowed values are [-1-511].

        Returns:
            The dynamic jump target associated to the event pattern.
        """
        return self._scpi_transport.ask(f"AWGC:EVEN:DJUM:DEF {pattern},{jump_target}").strip()

    @rpc_method
    def get_event_jump_mode(self) -> str:
        """Get event jump mode.

        Returns:
            The event jump mode.
        """
        return self._scpi_transport.ask("AWGC:EVEN:JMOD?").strip()

    @rpc_method
    def set_event_jump_mode(self, jump_mode: str) -> None:
        """Set the event jump mode.

        Parameters:
            jump_mode: The event jump mode, e.g. DJUMP, EJUMP.
        """
        self._scpi_transport.write(f"AWGC:EVEN:JMOD {jump_mode}")

    @rpc_method
    def get_event_jump_timing_mode(self) -> str:
        """Get event jump timing mode.

        Returns:
            Event jump timing mode.
        """
        return self._scpi_transport.ask("EVEN:JTIM?").strip()

    @rpc_method
    def set_event_jump_timing_mode(self, mode: str) -> None:
        """Set event jump timing mode to synchronous or to asynchronous.

        Parameters:
            Event jump timing mode. Valid modes are "SYNC" and "ASYN"
        """
        self._scpi_transport.write(f"EVEN:JTIM {mode}")

    @rpc_method
    def get_event_input_impedance(self) -> int:
        """Get external event input impedance.

        Returns:
            Impedance in Ohm. Possible values are 50 and 1000.
        """
        return int(float(self._scpi_transport.ask("EVEN:IMP?")))

    @rpc_method
    def set_event_input_impedance(self, impedance: int) -> None:
        """Set external event input impedance.

        Parameters:
            Impedance in Ohm. Possible values are 50 and 1000.
        """
        if impedance not in [50, 1000]:
            raise ValueError(f"Invalid impedance value {impedance}. Must be either 50 or 1000.")

        self._scpi_transport.write(f"EVEN:IMP {impedance}")

    @rpc_method
    def get_event_input_polarity(self) -> str:
        """Get event input polarity sign.

        Returns:
            Event input polarity sign as "POS" or "NEG".
        """
        return self._scpi_transport.ask("EVEN:POL?").strip()

    @rpc_method
    def set_event_input_polarity(self, polarity: str) -> None:
        """Set event input polarity sign.

        Parameters:
            Event input polarity sign as string "POS" or "POSitive" for positive polarity and "NEG" or "NEGative" for
            negative polarity.
        """
        if polarity.upper() in ["POS", "POSITIVE"]:
            target_polarity = "POSitive"

        elif polarity.upper() in ["NEG", "NEGATIVE"]:
            target_polarity = "NEGative"

        else:
            raise ValueError(f"Invalid polarity sign {polarity}")

        self._scpi_transport.write(f"EVEN:POL {target_polarity}")

    @rpc_method
    def get_event_level(self) -> float:
        """Get the event level voltage.

        Returns:
            The event level in Volts.
        """
        return float(self._scpi_transport.ask("EVEN:LEV?"))

    @rpc_method
    def set_event_level(self, level: float) -> None:
        """Set the event level voltage.

        Parameters:
            level: The target level voltage.
        """
        self._scpi_transport.write(f"EVEN:LEV {level:.6f}V")

    @rpc_method
    def get_marker_low(self, channel: int, marker: int) -> float:
        """Get the low level for a marker of a specific source channel.

        Parameters:
            channel: The channel number [1-4].
            marker: The marker number [1-2].

        Returns:
            The low level in Volts.
        """
        self._check_channel_number(channel)
        return float(self._scpi_transport.ask(f"SOUR{channel}:MARK{marker}:VOLT:LEV:IMM:LOW?"))

    @rpc_method
    def set_marker_low(self, channel: int, marker: int, low_level: float) -> None:
        """Set the low level for a marker of a specific source channel.

        Parameters:
            channel: The channel number [1-4].
            marker: The marker number [1-2].
            low_level: The low level in Volts.
        """
        self._check_channel_number(channel)
        self._scpi_transport.write(f"SOUR{channel}:MARK{marker}:VOLT:LEV:IMM:LOW {low_level:.3f}")

    @rpc_method
    def get_marker_high(self, channel: int, marker: int) -> float:
        """Get the high level for a marker of a specific source channel.

        Parameters:
            channel: The channel number [1-4].
            marker: The marker number [1-2].

        Returns:
            The high level in Volts.
        """
        self._check_channel_number(channel)
        return float(self._scpi_transport.ask(f"SOUR{channel}:MARK{marker}:VOLT:LEV:IMM:HIGH?"))

    @rpc_method
    def set_marker_high(self, channel: int, marker: int, high_level: float) -> None:
        """Set the high level for a marker of a specific source channel.

        Parameters:
            channel: The channel number [1-4].
            marker: The marker number [1-2].
            high_level: The high level in Volts.
        """
        self._check_channel_number(channel)
        self._scpi_transport.write(f"SOUR{channel}:MARK{marker}:VOLT:LEV:IMM:HIGH {high_level:.3f}")

    @rpc_method
    def get_marker_delay(self, channel: int, marker: int) -> float:
        """Query a marker delay of a specific source channel.

        Parameters:
            channel: The channel number [1-4].
            marker: The marker number [1-2].

        Returns:
            The delay in seconds.
        """
        self._check_channel_number(channel)
        return float(self._scpi_transport.ask(f"SOUR{channel}:MARK{marker}:DEL?"))

    @rpc_method
    def set_marker_delay(self, channel: int, marker: int, delay: float) -> None:
        """Set a marker delay of a specific source channel.

        Parameters:
            channel: The channel number [1-4].
            marker: The marker number [1-2].
            delay: The delay in seconds.
        """
        self._check_channel_number(channel)
        self._scpi_transport.write(f"SOUR{channel}:MARK{marker}:DEL {delay}")



"""Utility functions for QMI."""

import re
import threading

from typing import Tuple


def is_valid_object_name(name: str) -> bool:
    """Check that the specified name is an acceptable name for QMI.

    This function is used to check names for QMI contexts,
    QMI RPC objects, QMI instruments and QMI signals.

    Valid names contain at least 1 and at most 63 characters
    and consist of only letters, digits or the characters ``- _ ( )``.

    (Internally, QMI may use names which do not meet these criteria.)

    Parameters:
        name: Name to be validated.

    Returns:
        True if the named is valid, False if it is not valid.
    """

    if len(name) > 63:
        return False
    if not re.match(r"^[-_a-zA-Z0-9()]+$", name):
        return False
    return True


def format_address_and_port(address: Tuple[str, int]) -> str:
    """Format a host address and port number as a string ``"<host>:<port>"``.

    The host address may either be an IP address or a host name.

    If an IPv6 address is used, the formatting will add square brackets
    around the host address to separate it from the port number
    (see also RFC 3986, section 3.2.2).

    Parameters:
        address: Tuple (host, port).

    Returns:
        String containing the formatted address and port.
    """

    (host, port) = address
    if ':' in host:
        host = '[' + host + ']'
    return host + ':' + str(port)


def parse_address_and_port(address: str) -> Tuple[str, int]:
    """Parse host address and port number.

    The host address may either be an IP address or a host name.

    Parameters:
        address: String in format ``"<host>:<port>"``
            where `<host>` is either a host name or an IPv4 or IPv6 address,
            and `<port>` is a decimal TCP/UDP port number.
            If `<host>` is an IPv6 address, it must be enclosed in square brackets
            (e.g. "[::1]:5001") to avoid ambiguous interpretation of the ``:`` symbol.

    Returns:
        Tuple (host, port).

    Raises:
        ValueError: If an invalid address format is detected.
    """

    parts = address.rsplit(":", 1)
    if len(parts) != 2:
        raise ValueError("Invalid address format, expecting 'host:port'")
    (host, port_str) = parts

    if host.startswith("[") and host.endswith("]"):
        host = host[1:-1]
    port = int(port_str)

    return (host, port)


class AtomicCounter:
    """Simple atomic counter."""

    def __init__(self, initial_value: int = 0) -> None:
        self.lock = threading.Lock()
        self.count = initial_value

    def inc(self) -> None:
        """Increase the counter value by 1."""
        with self.lock:
            self.count += 1

    def dec(self) -> None:
        """Decrease the counter value by 1."""
        with self.lock:
            self.count -= 1

    def value(self) -> int:
        """Return the current counter value."""
        with self.lock:
            return self.count



"""
Measurement Computing Corporation (MCC), DAQ devices.

The qmi.instruments.mcc package provides limited support for:
- USB-1808 series devices.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.mcc.usb1808x import MCC_USB1808X as MCC_Usb1808x



import sys
import unittest.mock
sys.modules["usb.core"] = unittest.mock.Mock()
sys.modules["usb.core.find"] = unittest.mock.Mock()

from qmi.core.exceptions import QMI_TimeoutException
# For not causing an error by missing pyvisa library in the pipeline tests, we mock also the pyvisa existence
import tests.core.pyvisa_stub
sys.modules["pyvisa"] = tests.core.pyvisa_stub
sys.modules["pyvisa.errors"] = tests.core.pyvisa_stub.errors
with unittest.mock.patch("sys.platform", "win32"):
    from qmi.core.transport_gpib_visa import QMI_VisaGpibTransport

patcher = unittest.mock.patch("pyvisa.errors")


class TestQmiVisaGpibTransport(unittest.TestCase):
    """Test QMI_VisaGpibTransport."""

    def test_create_default_object(self):
        """Test instantiating the class with default values."""
        primary_addr = 1
        default_secondary_addr = None
        default_board = None
        default_timeout = 30
        expected_dev_str = f"QMI_VisaGpibTransport GPIB::{primary_addr}::INSTR"

        dev = QMI_VisaGpibTransport(primary_addr)
        dev_str = str(dev)

        self.assertIsNotNone(dev)
        self.assertEqual(expected_dev_str, dev_str)
        self.assertEqual(primary_addr, dev._primary_addr)
        self.assertEqual(default_board, dev._board)
        self.assertEqual(default_secondary_addr, dev._secondary_addr)
        self.assertEqual(default_timeout, dev._connect_timeout)

    def test_create_object(self):
        """Test instantiating the class with custom values."""
        primary_addr = 1
        board = 0
        secondnr = 2
        timeout = 0.1
        expected_dev_str = f"QMI_VisaGpibTransport GPIB{board}::{primary_addr}::{secondnr}::INSTR"

        dev = QMI_VisaGpibTransport(primary_addr, board=board, secondary_addr=secondnr, connect_timeout=timeout)
        dev_str = str(dev)

        self.assertIsNotNone(dev)
        self.assertEqual(expected_dev_str, dev_str)
        self.assertEqual(primary_addr, dev._primary_addr)
        self.assertEqual(board, dev._board)
        self.assertEqual(secondnr, dev._secondary_addr)
        self.assertEqual(timeout, dev._connect_timeout)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_open_close(self, mock):
        """Test opening and closing the instrument."""
        primary_nr = 1
        default_timeout = 30 * 1000
        terminations = '\n'
        dev = QMI_VisaGpibTransport(primary_nr)
        dev.open()
        dev.close()

        mock.assert_called_once()
        tests.core.pyvisa_stub.ResourceManager.open_resource.assert_called_once_with(
            f"GPIB::{primary_nr}::INSTR",
            open_timeout=default_timeout,
            write_termination=terminations,
            read_termination=terminations
        )

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_write(self, mock):
        """Test writing some data to the instrument."""
        data = bytes("data".encode())

        dev = QMI_VisaGpibTransport(1)
        dev.open()
        dev.write(data)
        dev.close()

        mock().write_raw.assert_called_once_with(data)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_read(self, mock):
        """Test reading something from the instrument."""
        expected = b"0"

        mock().read_raw.return_value = expected
        dev = QMI_VisaGpibTransport(1)
        dev.open()
        ret = dev.read(1, timeout=0.001)
        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(ret, expected)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    @unittest.mock.patch("pyvisa.errors.VisaIOError", new_callable=lambda: tests.core.pyvisa_stub.VisaIOError)
    def test_read_raises_exception(self, exc_mock, mock):
        """A test to see that read raises QMI_TimeoutException """
        # Arrange
        timeout_error = tests.core.pyvisa_stub.VI_ERROR_TMO
        mock().read_raw.side_effect = exc_mock(timeout_error)
        expected = QMI_TimeoutException(str(tests.core.pyvisa_stub.VisaIOError(timeout_error)))
        dev = QMI_VisaGpibTransport(1)
        dev.open()
        with self.assertRaises(QMI_TimeoutException) as exc:
            dev.read(1, timeout=0.001)

        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(repr(exc.exception), repr(expected))

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_read_until(self, mock):
        """This should forward the executions to read_until_timeout, so we test both in one go."""
        expected = b"01234"

        mock().read_raw.return_value = expected
        dev = QMI_VisaGpibTransport(1)
        dev.open()
        ret = dev.read_until(b"T3000", timeout=0.001)
        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(ret, expected)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    @unittest.mock.patch("pyvisa.errors.VisaIOError", new_callable=lambda: tests.core.pyvisa_stub.VisaIOError)
    def test_read_until_timeout_does_not_raise_timeout_exception(self, exc_mock, mock):
        """This tests that read_until_timeout catches timeout exception and returns empty bytes."""
        timeout_error = tests.core.pyvisa_stub.VI_ERROR_TMO
        expected = b""
        mock().read_raw.side_effect = exc_mock(timeout_error)

        dev = QMI_VisaGpibTransport(1)
        dev.open()
        ret = dev.read_until_timeout(10, timeout=0.001)
        dev.close()

        mock().read_raw.assert_called_once()
        self.assertEqual(ret, expected)

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    def test_discard_read(self, mock):
        """See that discard_read tries to read whatever is in the incoming instrument buffer.
        """
        expected = b"01234"

        mock().read_raw.return_value = expected
        dev = QMI_VisaGpibTransport(1)
        dev.open()
        # Read should empty the read buffer
        dev.discard_read()

        dev.close()

        mock().read_raw.assert_called_once()

    @unittest.mock.patch("pyvisa.ResourceManager.open_resource")
    @unittest.mock.patch("pyvisa.errors.VisaIOError", new_callable=lambda: tests.core.pyvisa_stub.VisaIOError)
    def test_discard_read_does_not_except(self, exc_mock, mock):
        """See that discard_read passes QMI_TimeoutException if read buffer and incoming instrument buffer are empty."""
        timeout_error = tests.core.pyvisa_stub.VI_ERROR_TMO
        mock().read_raw.side_effect = exc_mock(timeout_error)
        dev = QMI_VisaGpibTransport(1)
        dev.open()
        dev.discard_read()
        dev.close()

        mock().read_raw.assert_called_once()


if __name__ == "__main__":
    unittest.main()



"""Context managers for QMI RPC protocol contexts."""
from contextlib import contextmanager
from typing import Iterator, Optional, Protocol, TypeVar
import warnings

from qmi.core.pubsub import QMI_SignalReceiver, QMI_SignalSubscriber


class StartStoppable(Protocol):
    """Protocol for start_stop context class"""
    def start(self, *args, **kwargs) -> None: ...
    def stop(self) -> None: ...


class StartStopJoinable(StartStoppable, Protocol):
    """Protocol for start_stop context class, but has join as well."""
    def start(self, *args, **kwargs) -> None: ...
    def stop(self) -> None: ...
    def join(self) -> None: ...


class OpenClosable(Protocol):
    """Protocol for open_close context class. None of the `open` implementations take arguments."""
    def open(self) -> None: ...
    def close(self) -> None: ...


class LockUnlockable(Protocol):
    """Protocol for lock_unlock context class. None of the `lock` implementations take arguments."""
    def lock(self, *args, **kwargs) -> None: ...
    def unlock(self, *args, **kwargs) -> None: ...


_SS = TypeVar("_SS", bound=StartStoppable)
_SSJ = TypeVar("_SSJ", bound=StartStopJoinable)
_OC = TypeVar("_OC", bound=OpenClosable)
_LU = TypeVar("_LU", bound=LockUnlockable)


@contextmanager
def start_stop(thing: _SS, *args, **kwargs) -> Iterator[_SS]:
    thing.start(*args, **kwargs)
    try:
        yield thing
    finally:
        thing.stop()


@contextmanager
def start_stop_join(thing: _SSJ, *args, **kwargs) -> Iterator[_SSJ]:
    warnings.warn(
        "This context manager is obsoleted. The tasks can now by managed directly by their own context manager "
        "by calling `with qmi.make_task('task_name', TaskClass, args, kwargs) as task: ...`.",
        DeprecationWarning,
        stacklevel=3
    )
    thing.start(*args, **kwargs)
    try:
        yield thing
    finally:
        thing.stop()
        thing.join()


@contextmanager
def open_close(thing: _OC) -> Iterator[_OC]:
    warnings.warn(
        "This context manager is obsoleted. The instruments can now by managed directly by their own context manager "
        "by calling `with qmi.make_instrument('instr_name', InstrClass, args, kwargs) as instr: ...`.",
        DeprecationWarning,
        stacklevel=3
    )
    thing.open()
    try:
        yield thing
    finally:
        thing.close()


@contextmanager
def lock_unlock(thing: _LU, *args, **kwargs) -> Iterator[_LU]:
    thing.lock(*args, **kwargs)
    try:
        yield thing
    finally:
        thing.unlock()


@contextmanager
def subscribe_unsubscribe(
        signal: QMI_SignalSubscriber, receiver: Optional[QMI_SignalReceiver]
    ) -> Iterator[QMI_SignalReceiver]:
    receiver = receiver if receiver is not None else QMI_SignalReceiver()
    signal.subscribe(receiver)
    try:
        yield receiver
    finally:
        signal.unsubscribe(receiver)




""" Instrument driver for the Raspberry Pi peripheral I/O.
"""

import enum
import typing

# Lazy import of the GPIO module. See the function _import_modules() below.
if typing.TYPE_CHECKING:
    import RPi.GPIO as GPIO
else:
    GPIO = None

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method


def _import_modules() -> None:
    """Import the Raspberry Pi GPIO module.

    This import is done in a function, instead of at the top-level,
    to avoid an unnecessary dependency for programs that do not access
    the instrument directly.
    """
    global GPIO
    if GPIO is None:
        import RPi.GPIO as GPIO  # pylint: disable=W0621

# See:
#
# https://pypi.org/project/RPi.GPIO/
# https://www.raspberrypi-spy.co.uk/2012/06/simple-guide-to-the-rpi-gpio-header-and-pins

# GPIO pins (BOARD numbering):
#
#     3, 5, 7, 8, 10, 11, 12, 13, 15, 16, 18, 19, 21, 22, 23, 24, 26, 29, 31, 32, 33, 35, 36, 37, 38, 40]


class RaspberryPiGPIO(QMI_Instrument):
    """Instrument driver for the Raspberry Pi I/O, based on the 'RPi' module.

    NOTE: It is only sensible to instantiate this driver on a Raspberry Pi.
    Of course, from then on, it can be accessed via RPC.

    This is based on the package RPi.GPIO. We expose a subset of its functionality.
    """
    # Fault codes returned by get_operating_fault().
    class gpio_mode(enum.Enum):
        INPUT = 0
        OUTPUT = 1

    def __init__(self, context: QMI_Context, name: str) -> None:
        super().__init__(context, name)

        _import_modules()

    @rpc_method
    def open(self) -> None:
        self._check_is_closed()
        GPIO.setmode(GPIO.BOARD)
        super().open()

    @rpc_method
    def close(self) -> None:
        self._check_is_open()
        GPIO.cleanup()
        super().close()

    @rpc_method
    def gpio_setup(self, pin_nr: int, mode: gpio_mode) -> None:
        """Setup a GPIO pin in given mode.

        Parameters:
            pin_nr: The pin number to setup.
            mode:   The mode the pin should be setup. Options: gpio_mode.INPUT, gpio_mode.OUTPUT.
        """
        self._check_is_open()
        mode = GPIO.IN if mode == RaspberryPiGPIO.gpio_mode.INPUT else GPIO.OUT
        GPIO.setup(pin_nr, mode)

    @rpc_method
    def gpio_input(self, pin_nr: int) -> bool:
        """Check if given pin is in high lor low state.

        Parameters:
            pin_nr: The pin number to check.

        Returns:
            value:  True if the pin state is GPIO.HIGH, False otherwise.
        """
        self._check_is_open()
        value = GPIO.input(pin_nr)
        value = (value == GPIO.HIGH)
        return value

    @rpc_method
    def gpio_output(self, pin_nr: int, value: bool) -> None:
        """Set given pin output to high or low.

        Parameters:
            pin_nr: The pin number to be set.
            value:  True to set pin to GPIO.HIGH or False to set pin to GPIO.LOW.
        """
        self._check_is_open()
        value = GPIO.HIGH if value else GPIO.LOW
        GPIO.output(pin_nr, value)



import unittest.mock

from qmi.instruments.picoquant.support._library_wrapper import _LibWrapper
from qmi.instruments.picoquant.support import _phlib_function_signatures, \
    _mhlib_function_signatures, _hhlib_function_signatures


class LibraryWrapperTestCase(unittest.TestCase):

    @unittest.mock.patch("sys.platform", "darwin")
    def test_unsupported_platform_fail(self):
        with self.assertRaises(OSError) as exc:
            _LibWrapper("HH")

        self.assertEqual(str(exc.exception), "Unsupported platform.")

    @unittest.mock.patch("sys.platform", "linux")
    def test_unsupported_harp_lin(self):
        false_lib = "ASDF"
        with self.assertRaises(FileNotFoundError) as exc:
            _LibWrapper(false_lib)

        self.assertEqual(str(exc.exception), "Unknown library: {lib}.".format(lib=false_lib))

    @unittest.mock.patch("sys.platform", "win32")
    def test_unsupported_harp_win(self):
        false_lib = "ASDF"
        with self.assertRaises(FileNotFoundError) as exc:
            _LibWrapper(false_lib)

        self.assertEqual(str(exc.exception), "Unknown library: {lib}.".format(lib=false_lib))

    @unittest.mock.patch("sys.platform", "linux")
    def test_multiharp_wrapper_annotations_lin(self):
        attr_list = [fs[0] for fs in _mhlib_function_signatures._mhlib_function_signatures]
        with unittest.mock.patch("ctypes.cdll"):
            lib_wrapper = _LibWrapper("MH")
            annotations = dir(lib_wrapper._lib)
            for a in attr_list:
                self.assertIn(a, annotations)

    @unittest.mock.patch("sys.platform", "linux")
    def test_hydraharp_wrapper_annotations_lin(self):
        attr_list = [fs[0] for fs in _hhlib_function_signatures._hhlib_function_signatures]
        with unittest.mock.patch("ctypes.cdll"):
            lib_wrapper = _LibWrapper("HH")
            annotations = dir(lib_wrapper._lib)
            for a in attr_list:
                self.assertIn(a, annotations)

    @unittest.mock.patch("sys.platform", "linux")
    def test_picoharp_wrapper_annotations_lin(self):
        attr_list = [fs[0] for fs in _phlib_function_signatures._phlib_function_signatures]
        with unittest.mock.patch("ctypes.cdll"):
            lib_wrapper = _LibWrapper("PH")
            annotations = dir(lib_wrapper._lib)
            for a in attr_list:
                self.assertIn(a, annotations)

    @unittest.mock.patch("sys.platform", "win32")
    def test_multiharp_wrapper_annotations_win(self):
        attr_list = [fs[0] for fs in _mhlib_function_signatures._mhlib_function_signatures]
        with unittest.mock.patch("ctypes.WinDLL", create=True):
            lib_wrapper = _LibWrapper("MH")
            annotations = dir(lib_wrapper._lib)
            for a in attr_list:
                self.assertIn(a, annotations)

    @unittest.mock.patch("sys.platform", "win32")
    def test_hydraharp_wrapper_annotations_win(self):
        attr_list = [fs[0] for fs in _hhlib_function_signatures._hhlib_function_signatures]
        with unittest.mock.patch("ctypes.WinDLL", create=True):
            lib_wrapper = _LibWrapper("HH")
            annotations = dir(lib_wrapper._lib)
            for a in attr_list:
                self.assertIn(a, annotations)

    @unittest.mock.patch("sys.platform", "win32")
    def test_picoharp_wrapper_annotations_win(self):
        attr_list = [fs[0] for fs in _phlib_function_signatures._phlib_function_signatures]
        with unittest.mock.patch("ctypes.WinDLL", create=True):
            lib_wrapper = _LibWrapper("PH")
            annotations = dir(lib_wrapper._lib)
            for a in attr_list:
                self.assertIn(a, annotations)

    @unittest.mock.patch("sys.platform", "win32")
    def test_wrapper_exceptions_32bit_win(self):
        """Test that the library wrapper will try to call also for the 32-bit library if 64-bit library call excepts."""
        with unittest.mock.patch("ctypes.WinDLL", create=True, side_effect=[BaseException("No lib"), None]) \
                as ptc:

            _LibWrapper("MH")

        ptc.assert_any_call("mhlib64.dll")
        ptc.assert_called_with("mhlib.dll")

        with unittest.mock.patch("ctypes.WinDLL", create=True, side_effect=[BaseException("No lib"), None]) \
                as ptc:

            _LibWrapper("HH")

        ptc.assert_any_call("hhlib64.dll")
        ptc.assert_called_with("hhlib.dll")

        with unittest.mock.patch("ctypes.WinDLL", create=True, side_effect=[BaseException("No lib"), None]) \
                as ptc:

            _LibWrapper("PH")

        ptc.assert_any_call("phlib64.dll")
        ptc.assert_called_with("phlib.dll")


if __name__ == "__main__":
    unittest.main()



"""
Hübner Photonics, diode laser.

The qmi.instruments.cobolt package provides support for:
- Continuous Wave 06-01 series diode laser.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.cobolt.laser_06_01 import Cobolt_Laser_06_01 as Hubner_Cobolt0601



"""Instrument driver for the Thorlabs MFF101/MFF102 motorized flip mounts."""

import logging
import struct
import time
from typing import NamedTuple, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

# Hardware information returned by instrument.
_HwInfo = NamedTuple('_HwInfo', [
    ('serial_number', int),
    ('model_number', str),
    ('hw_type', int),
    ('fw_version', Tuple[int, int, int]),
    ('hw_version', int),
    ('mod_state', int),
    ('num_channels', int)
])


def _format_msg_hw_req_info(dest: int, source: int) -> bytes:
    """Prepare a HW_REQ_INFO message.

    The controller will answer with MSG_HW_GET_INFO.

    Parameters:
        dest: Destination address.
        source: Source address.
    """
    assert 0 <= dest < 128
    assert 0 <= source < 128
    msg = bytes([0x05, 0x00, 0x00, 0x00, dest, source])
    return msg


def _format_msg_mot_req_statusbits(chan_iden: int, dest: int, source: int) -> bytes:
    """Prepare a MOT_REQ_STATUSBITS message.

    The controller will answer with MSG_MOT_GET_STATUSBITS.

    Parameters:
        chan_iden: Channel identifier.
        dest: Destination address.
        source: Source address.
    """
    assert 0 <= dest < 128
    assert 0 <= source < 128
    msg = bytes([0x29, 0x04, chan_iden, 0x00, dest, source])
    return msg


def _format_msg_mot_move_jog(chan_iden: int, direction: int, dest: int, source: int) -> bytes:
    """Prepare a MOT_MOVE_JOG message.

    The controller will answer with MSG_MOT_MOVE_COMPLETED.

    Parameters:
        chan_iden: Channel identifier.
        direction: Direction to move (0x01=forward, 0x02=reverse).
        dest: Destination address.
        source: Source address.
    """
    assert 0 <= chan_iden <= 255
    assert 0 <= direction <= 255
    assert 0 <= dest < 128
    assert 0 <= source < 128
    msg = bytes([0x6A, 0x04, chan_iden, direction, dest, source])
    return msg


def _parse_msg_hw_get_info(msg: bytes) -> _HwInfo:
    assert len(msg) >= 6
    (message_id,) = struct.unpack_from("<H", msg, 0)
    if message_id != 0x0006:
        raise QMI_InstrumentException("Got message ID {:04x} while expecting MSG_HW_GET_INFO".format(message_id))

    if len(msg) < 90:
        raise QMI_InstrumentException("Got {} bytes while expecting 90 bytes MSG_HW_GET_INFO".format(len(msg)))

    (serial_number,) = struct.unpack_from("<I", msg, 6)
    model_number = msg[10:18].rstrip(b"\x00").decode('ascii', errors='replace')
    (hw_type,) = struct.unpack_from("<H", msg, 18)
    fw_version = (msg[22], msg[21], msg[20])
    (hw_version, mod_state, num_channels) = struct.unpack_from("<HHH", msg, 84)

    return _HwInfo(serial_number, model_number, hw_type, fw_version, hw_version, mod_state, num_channels)


def _parse_msg_mot_get_statusbits(msg: bytes) -> int:
    assert len(msg) >= 6
    (message_id,) = struct.unpack_from("<H", msg, 0)
    if message_id != 0x042A:
        raise QMI_InstrumentException("Got message ID {:04x} while expecting MSG_MOT_GET_STATUSBITS".format(message_id))

    if len(msg) < 12:
        raise QMI_InstrumentException("Got {} bytes while expecting 12 bytes MSG_MOT_GET_STATUSBITS".format(len(msg)))

    (statusbits,) = struct.unpack_from("<I", msg, 8)
    return statusbits


def _parse_msg_mot_move_completed(msg: bytes) -> None:
    assert len(msg) >= 6
    (message_id,) = struct.unpack_from("<H", msg, 0)
    if message_id != 0x0464:
        raise QMI_InstrumentException("Got message ID {:04x} while expecting MSG_MOT_MOVE_COMPLETED".format(message_id))


class Thorlabs_MFF10X(QMI_Instrument):
    """Instrument driver for the Thorlabs MFF101/MFF102 motorized flip mounts."""

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str
                 ) -> None:
        """Initialize driver.

        The flipmount presents itself as an USB serial port.
        The transport descriptor should refer to the serial port device,
        e.g. "serial:/dev/ttyUSB1"

        Parameters:
            name: Name for this instrument instance.
            transport: Transport descriptor to access the instrument.
        """
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"baudrate": 115200, "rtscts": True})

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._transport.open()
        self._transport.discard_read()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    def _read_message(self, timeout: float) -> bytes:
        """Read a binary message from the instrument."""

        assert self._transport is not None
        msg = self._transport.read(nbytes=6, timeout=timeout)

        if (msg[4] & 0x80) != 0:
            # extended message
            (datalen,) = struct.unpack_from("<H", msg, 2)
            if datalen > 255:
                raise QMI_InstrumentException("Received invalid message header from instrument")
            payload = self._transport.read(nbytes=datalen, timeout=timeout)
            msg = msg + payload

        return msg

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""

        self._check_is_open()
        assert self._transport is not None

        req_msg = _format_msg_hw_req_info(dest=0x11, source=0x01)
        self._transport.discard_read()
        self._transport.write(req_msg)

        answer_msg = self._read_message(timeout=1.0)
        hw_info = _parse_msg_hw_get_info(answer_msg)

        return QMI_InstrumentIdentification(vendor="Thorlabs",
                                            model=hw_info.model_number,
                                            serial=str(hw_info.serial_number),
                                            version="{}.{}.{}".format(*hw_info.fw_version))

    @rpc_method
    def get_position(self) -> int:
        """Get current position of flip mount.

        Returns:
            Current position:
                0 = mount in undefined position;
                1 = mount in forward position;
                2 = mount in reverse position.
        """

        self._check_is_open()
        assert self._transport is not None

        req_msg = _format_msg_mot_req_statusbits(chan_iden=1, dest=0x11, source=0x01)
        self._transport.discard_read()
        self._transport.write(req_msg)

        answer_msg = self._read_message(timeout=1.0)
        statusbits = _parse_msg_mot_get_statusbits(answer_msg)

        return statusbits & 3

    @rpc_method
    def move_mount(self,
                   direction: int,
                   wait_complete: bool,
                   timeout: float = 5.0
                   ) -> None:
        """Move the flip mount and wait until move is finished.

        Parameters:
            direction: Direction to move:
                1 = move to forward position;
                2 = move to reverse position.
            wait_complete: True to wait until the move completes.
            timeout: Maximum time to wait for response (seconds).

        Raises:
            ~qmi.core.exceptions.QMI_TimeoutException: If the move does not complete within the specified timeout.
        """

        self._check_is_open()
        assert self._transport is not None

        if direction not in (1, 2):
            raise ValueError("Invalid value for parameter 'direction' (must be 1 or 2).")

        req_msg = _format_msg_mot_move_jog(chan_iden=1, direction=direction, dest=0x11, source=0x01)
        self._transport.discard_read()
        self._transport.write(req_msg)

        if wait_complete:
            waited = 0.0
            completed = False
            interval = 0.05
            while (not completed) and (waited < timeout):
                time.sleep(interval)
                waited += interval
                if self.get_position() == direction:
                    completed = True
            if not completed:
                raise QMI_TimeoutException("Timeout while waiting for moving flipmount")



""" This module contains the base class for QMI instrument driver for the Picoquant 'Harp instruments."""

import ctypes
import enum
import logging
import time
from enum import Enum
from threading import Lock
from typing import Optional, Dict, List, Tuple, TypeVar, Type

import numpy as np

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InvalidOperationException, QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument
from qmi.core.messaging import _PeerTcpConnection
from qmi.core.pubsub import QMI_Signal
from qmi.core.rpc import rpc_method
from qmi.instruments.picoquant import EventFilterMode, SYNC_TYPE
from qmi.instruments.picoquant.support._realtime import RealTimeHistogram, RealTimeCountRate, NUM_CHANNELS
from qmi.instruments.picoquant.support._events import _FetchEventsThread, _MODE
from qmi.instruments.picoquant.support._library_wrapper import _LibWrapper

_logger = logging.getLogger(__name__)

_ENUM_T = TypeVar('_ENUM_T', bound=Enum)
MAX_MESSAGE_SIZE = _PeerTcpConnection.MAX_MESSAGE_SIZE


def _str_to_enum(enum_type: Type[_ENUM_T], str_value: str) -> _ENUM_T:
    """Convert a string value to an Enum-type value, by comparing it to the enum value names."""
    try:
        return enum_type[str_value]
    except KeyError:
        allowed_values = tuple(ev.name for ev in enum_type)
        raise ValueError("Bad value {!r}, expected one of {!r}".format(str_value, allowed_values))


@enum.unique
class _EDGE(enum.Enum):
    """Symbolic constants for routines that take an `edge` argument:

    * Function :func:`~MultiHarpDevice.setMeasurementControl`, `startedge` and `stopedge` arguments;
    * Function :func:`~MultiHarpDevice.setSyncEdgeTrigger`, `edge` argument;
    * Function :func:`~MultiHarpDevice.setInputEdgeTrigger`, `edge` argument.

    These are defined as preprocessor symbols in the ``mhdefin.h`` C header file.
    """
    RISING = 1
    """Rising edge."""
    FALLING = 0
    """Falling edge."""


@enum.unique
class _FEATURE(enum.IntFlag):
    """Bitfield constants for the return value of the :func:`~MultiHarpDevice.getFeatures` function.

    These are defined as preprocessor symbols in the ``mhdefin.h`` C header file.
    """
    DLL = 0x0001
    """DLL License available."""
    TTTR = 0x0002
    """TTTR mode available."""
    MARKERS = 0x0004
    """Markers available."""
    LOWRES = 0x0008
    """Long range mode available"""
    TRIGOUT = 0x0010
    """Trigger output available."""
    PROG_TD = 0x0020
    """Programmable deadtime available."""


class _PicoquantHarp(QMI_Instrument):

    # Some functions have model-dependent variable/string differences. Use this to distinguish
    _MODEL: str = ""

    # The QMI RPC mechanism can not handle messages larger than what is set in qmi.core.messaging.
    # To avoid RPC errors, we limit the number of events returned per call.
    MAX_EVENTS_PER_CALL = MAX_MESSAGE_SIZE

    # Signal published to report real-time histograms based on T2 event data.
    sig_histogram = QMI_Signal([RealTimeHistogram])

    # Signal published to report real-time count rates based on T2 event data.
    sig_countrate = QMI_Signal([RealTimeCountRate])

    def __init__(self, context: QMI_Context, name: str, serial_number: str, max_pending_events: int = 10 ** 8) -> None:
        """Instantiate the instrument driver. This is the base class for all *Harp instruments.

        Parameters:
            context: the QMI_Context that manages us
            name: the name of the instrument instance
            serial_number: the serial number of the instrument to be opened.
            max_pending_events: Only Relevant for T2 capturing. Defaults to 10e8 events.
        """
        super().__init__(context, name)
        self._serial_number = serial_number
        self._max_pending_events = max_pending_events
        self._lazy_lib: Optional[_LibWrapper] = None
        self._devidx = -1
        self._lib_version: str = ""
        self._fetch_events_thread: Optional[_FetchEventsThread] = None
        self._mode: Optional[_MODE] = None
        self._measurement_running = False
        self._measurement_start_time = 0.0
        self._event_filter_channels: Dict[int, EventFilterMode] = {}
        self._event_filter_aperture = (0, 0)
        self._actuallen = 65536  # Actual histogram length value

        # The instrument may be accessed concurrently by two threads:
        #  - public API functions are called through the RPC thread that manages this instrument instance
        #  - the function _read_fifo() is called by the background event fetching thread.
        # For this reason, interaction with the instrument must be protected with an explicit mutex.
        self._device_lock = Lock()

    @property
    def _max_dev_num(self):
        raise NotImplementedError()

    @property
    def _ttreadmax(self):
        raise NotImplementedError()

    @property
    def _model(self):
        raise NotImplementedError()

    @property
    def _lib(self) -> _LibWrapper:
        raise NotImplementedError

    def _read_fifo(self) -> np.ndarray:
        """Read event FIFO data.

        This is an internal method. It gets called by the background thread when measuring in T2 mode.

        CPU time during wait for completion will be yielded to other processes/threads. The call will return after a
        timeout period of approximately 1 ms if no more data could be fetched. The actual time to return may vary
        towards 2..3 ms due to USB overhead and operating system latencies.

        Returns:
            Raw FIFO data as a 1-dimensional array of 32-bit unsigned integers.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            fifo_data = np.empty(self._ttreadmax, dtype=np.uint32)
            buffer = fifo_data.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32))
            n_actual = ctypes.c_int()
            if self._model == "MH":
                self._lib.ReadFiFo(self._devidx, buffer, n_actual)

            else:
                self._lib.ReadFiFo(self._devidx, buffer, self._ttreadmax, n_actual)

            n_actual_int = n_actual.value
            return fifo_data[:n_actual_int].copy()

    @rpc_method
    def open(self) -> None:
        """Open the device.

        This method also starts a background thread which will be responsible
        for fetching event data from the instrument.
        """

        self._check_is_closed()
        _logger.info("[%s] Opening connection to instrument", self._name)

        vers = ctypes.create_string_buffer(8)
        self._lib.GetLibraryVersion(vers)
        self._lib_version = vers.value.decode()
        _logger.debug("Library version = %s", self._lib_version)

        with self._device_lock:
            for devidx in range(self._max_dev_num):
                try:
                    serial = ctypes.create_string_buffer(8)
                    self._lib.OpenDevice(devidx, serial)
                    decoded_serial = serial.value.decode()
                    if not decoded_serial:
                        try:
                            if self._model in ["MH", "HH"]:
                                self._lib.Initialize(devidx, 0, 0)  # HydraHarp and MultiHarp

                            elif self._model in ["PH", "TH260"]:
                                self._lib.Initialize(devidx, 0)  # PicoHarp and TimeHarp

                            else:
                                raise NotImplementedError(f"Model {self._model} is not implemented!")

                        except QMI_InstrumentException:
                            continue

                        serial = ctypes.create_string_buffer(8)
                        self._lib.GetSerialNumber(devidx, serial)
                        decoded_serial = serial.value.decode()

                    if decoded_serial == self._serial_number:
                        self._devidx = devidx
                        break

                    self._lib.CloseDevice(devidx)

                except QMI_InstrumentException:
                    continue

            if self._devidx == -1:
                raise QMI_InstrumentException(f"No device with serial number {self._serial_number!r} found.")

        self._fetch_events_thread = _FetchEventsThread(self._read_fifo,
                                                       self.sig_histogram.publish,  # type: ignore
                                                       self.sig_countrate.publish,  # type: ignore
                                                       self._max_pending_events)
        self._fetch_events_thread.start()

        super().open()

    @rpc_method
    def close(self) -> None:
        """Close the device.

        Stops the background event fetching thread and closes the underlying device.
        """

        self._check_is_open()

        assert (self._fetch_events_thread is not None)
        self._fetch_events_thread.shutdown()
        self._fetch_events_thread.join()
        self._fetch_events_thread = None

        _logger.info("[%s] Closing connection to instrument", self._name)
        with self._device_lock:
            self._devidx = -1
        self._mode = None
        super().close()

    @rpc_method
    def get_error_string(self, error_code: int) -> str:
        """ Get the error string for an error code number.

        Inputs:
            error_code: The error code number to be translated

        Returns:
            Error code's description string.
        """
        self._check_is_open()
        with self._device_lock:
            error_string = ctypes.create_string_buffer(40)
            self._lib.GetErrorString(error_string, error_code)

        _logger.error("[%s] Got error %s", self._name, error_string.value.decode())
        return error_string.value.decode()

    @rpc_method
    def get_hardware_info(self) -> Tuple[str, str, str]:
        """Get model, part nr, and version of the device.

        Note:
            This function can only be used after the :func:`initialize` method was successfully called.

        Returns:
            A 3-tuple of consisting of (model, partno, version) strings.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            model = ctypes.create_string_buffer(24)
            partno = ctypes.create_string_buffer(8)
            version = ctypes.create_string_buffer(8)
            self._lib.GetHardwareInfo(self._devidx, model, partno, version)
            decoded_model = model.value.decode()
            decoded_partno = partno.value.decode()
            decoded_version = version.value.decode()
            return decoded_model, decoded_partno, decoded_version

    @rpc_method
    def get_serial_number(self) -> str:
        """Get the device serial number.

        Note:
            This function can only be used after the :func:`initialize` method was successfully called.

        Returns:
            The serial number for example '1035683'.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            serial = ctypes.create_string_buffer(8)
            self._lib.GetSerialNumber(self._devidx, serial)
            return serial.value.decode()

    @rpc_method
    def get_features(self) -> List[str]:
        """Get device features.

        This function is usually not needed by the user. It is mainly for integration in PicoQuant system software
        such as SymPhoTime in order to figure out in a standardized way the capabilities the device has.

        Note:
            This function can only be used after the :func:`initialize` method was successfully called.

        Returns:
            A list of strings, each describing a feature of the device.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            features_bitset = ctypes.c_int()
            self._lib.GetFeatures(self._devidx, features_bitset)
            features = _FEATURE(features_bitset.value)
            return [feature.name for feature in _FEATURE if feature in features and feature.name is not None]

    @rpc_method
    def get_base_resolution(self) -> Tuple[float, int]:
        """Get the resolution and binsteps of the device.

        Returns:
            A tuple `(resolution, binsteps)` where resolution (float) is the base resolution in
            picoseconds; binsteps (int) is the number of allowed binning steps.
        """
        self._check_is_open()
        with self._device_lock:
            resolution = ctypes.c_double()
            binsteps = ctypes.c_int()
            self._lib.GetBaseResolution(self._devidx, resolution, binsteps)
            return resolution.value, binsteps.value

    @rpc_method
    def get_debug_info(self) -> str:
        """Get debugging info of the device.

        Call this immediately after receiving an error from any library call or after detecting a `SYSERROR` from
        `get_flags`. In case of `SYSERROR`, please provide this information to support.

        Returns:
            A debug info string (65,655 characters, max).

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            debuginfo = ctypes.create_string_buffer(65536)
            if self._model == "MH":
                self._lib.GetDebugInfo(self._devidx, debuginfo)

            else:
                self._lib.GetHardwareDebugInfo(self._devidx, debuginfo)

            return debuginfo.value.decode()

    @rpc_method
    def get_number_of_input_channels(self) -> int:
        """Get number of input channels of the device.

        Note:
            This function can only be used after the :func:`initialize` method was successfully called.

        Returns:
            Number of input channels.
            When a channel index is passed to another method of this driver, the valid range
            of channel index is from 0 to `number_of_channels` - 1.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            nchannels = ctypes.c_int()
            self._lib.GetNumOfInputChannels(self._devidx, nchannels)
            return nchannels.value

    @rpc_method
    def calibrate(self) -> None:
        """Calibrate the <xxx>Harp instrument. <xxx> = Hydra, Pico

        This method should be called before starting a measurement.

        Calibrates the time measurement circuits. This is necessary whenever temperature changes > 5 K occurred.
        Calibrate after warming up of the instrument before starting serious measurements. Warming up takes about 20
        to 30 minutes dependent on the lab temperature. It is completed when the cooling fan starts to run for the first
        time. Calibration takes only a few seconds. Calibration is only important for serious measurements.
        You can use the instrument during the warming-up period for set–up and preliminary measurements. For very long
        measurements, allow some more time for thermal stabilization, calibrate immediately before the measurement
        commences and try to maintain a stable room temperature during the measurement. The permissible ambient
        temperature is 15°C to 35 °C. Do not obstruct the cooling fan at the back and the air inlets at the bottom of
        the housing.
        """
        raise NotImplementedError()

    @rpc_method
    def set_marker_edges(self, me0_str: str, me1_str: str, me2_str: str, me3_str: str) -> None:
        """Set marker edges.

        Change the active edge on which the external TTL signals connected to the marker inputs are triggering.
        Only meaningful in TTTR mode.

        Parameters:
            me0_str (str): edge of marker signal 0.
            me1_str (str): edge of marker signal 1.
            me2_str (str): edge of marker signal 2.
            me3_str (str): edge of marker signal 3.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        me0 = _str_to_enum(_EDGE, me0_str)
        me1 = _str_to_enum(_EDGE, me1_str)
        me2 = _str_to_enum(_EDGE, me2_str)
        me3 = _str_to_enum(_EDGE, me3_str)
        with self._device_lock:
            self._lib.SetMarkerEdges(self._devidx, me0.value, me1.value, me2.value, me3.value)

    @rpc_method
    def set_marker_enable(self, en0: bool, en1: bool, en2: bool, en3: bool) -> None:
        """Set marker enable.

        Used to enable or disable the external TTL marker inputs.
        Only meaningful in TTTR mode.

        Parameters:
            en0: edge of marker signal 0.
            en1: edge of marker signal 1.
            en2: edge of marker signal 2.
            en3: edge of marker signal 3.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.SetMarkerEnable(self._devidx, int(en0), int(en1), int(en2), int(en3))

    @rpc_method
    def set_marker_holdoff_time(self, holdofftime: int) -> None:
        """Set marker hold-off time.

        Parameters:
            holdofftime: hold-off time in [ns]. Ranges from 0 ns 25500 ns.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.SetMarkerHoldoffTime(self._devidx, holdofftime)

    @rpc_method
    def start_measurement(self, tacq: int) -> None:
        """Start a measurement.

        Parameters:
            tacq: acquisition time, in [ms], ranging from 1 to 360000000; corresponding to 100 hours)

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        if self._measurement_running:
            raise QMI_InvalidOperationException("Measurement still active")

        assert self._fetch_events_thread is not None
        with self._device_lock:
            self._lib.StartMeas(self._devidx, tacq)

        self._measurement_start_time = time.time()
        self._measurement_running = True
        if self._mode == _MODE.T2:
            # Start collecting data in the background thread.
            self._fetch_events_thread.activate(_MODE.T2)

        elif self._mode == _MODE.T3:
            with self._device_lock:
                # Get sync rate and resolution from the Harp first
                sync_rate = ctypes.c_int()
                self._lib.GetSyncRate(self._devidx, sync_rate)
                resolution = ctypes.c_double()
                self._lib.GetResolution(self._devidx, resolution)

            # Start collecting data in the background thread.
            self._fetch_events_thread.activate(_MODE.T3, sync_rate.value, int(resolution.value))

        _logger.info("[%s] Activated events thread fetching in mode %r", self._name, self._mode)

    @rpc_method
    def stop_measurement(self) -> None:
        """Stop a running measurement.

        This call can be used to force a stop before the acquisition time expires.

        Important:
            For cleanup purposes, this must be called after a measurement, even if the measurement has expired on its
            own!

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        assert self._fetch_events_thread is not None
        with self._device_lock:
            self._lib.StopMeas(self._devidx)
        if self._mode in [_MODE.T2, _MODE.T3]:
            # Stop collecting data in the background thread.
            self._fetch_events_thread.deactivate()
            _logger.info("[%s] De-activated events thread fetching for mode %r", self._name, self._mode)
        self._measurement_running = False

    @rpc_method
    def get_measurement_active(self) -> bool:
        """Check if measurement is active.

        This method will return False if a measurement has expired on its own, even if
        the method `stop_measurement()` has not yet been called.

        Returns:
            | True - acquisition time still running.
            | False - acquisition time has ended.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            ctcstatus = ctypes.c_int()
            self._lib.CTCStatus(self._devidx, ctcstatus)
            ctcstatus_int = ctcstatus.value
        return ctcstatus_int == 0

    @rpc_method
    def get_measurement_start_time(self) -> float:
        """Return the approximate POSIX timestamp of the start of the current measurement.

        The returned timestamp is based on the computer system clock.
        Its accuracy depends on the latency of USB communication with the instrument,
        process scheduling delay and the accuracy of the system clock.

        Returns:
            POSIX timestamp as a floating point number
        """
        self._check_is_open()
        return self._measurement_start_time

    @rpc_method
    def get_events(self) -> np.ndarray:
        """Return events recorded by the instrument in T2 or T3 mode.

        While a measurement is active, a background thread continuously reads
        event records from the instrument and stores the events in a buffer.
        This method takes all pending events, removes them from the buffer and returns them.

        Events are returned as an array of event records.
        Each event record contains two fields:
        | `type` (uint8): The channel index where the event was recorded, or 64 for a SYNC event.
        | `timestamp` (uint64): Event timestamp as a multiple of the instrument base resolution.

        This method may only be used in T2 or T3 mode.

        Returns:
            Numpy array containing event records.
        """
        self._check_is_open()
        assert self._fetch_events_thread is not None
        _logger.info("[%s] Fetching events from thread in mode %r", self._name, self._mode)
        (_timestamp, events) = self._fetch_events_thread.get_events(self.MAX_EVENTS_PER_CALL)
        return events

    @rpc_method
    def get_timestamped_events(self) -> Tuple[float, np.ndarray]:
        """Return events recorded by the instrument in T2 or T3 mode.

        While a measurement is active, a background thread continuously reads
        event records from the instrument and stores the events in a buffer.
        This method takes all pending events, removes them from the buffer and returns them.

        Events are returned as an array of event records.
        Each event record contains two fields:
        | `type` (uint8): The channel index where the event was recorded, or 64 for a SYNC event.
        | `timestamp` (uint64): Event timestamp as a multiple of the instrument base resolution.

        This method may only be used in T2 or T3 mode.

        Returns:
            Tuple `(timestamp, events)`
            where `timestamp` is the approximate wall-clock time where the last event record was received,
            and `events` is a Numpy array containing the event records.
        """
        self._check_is_open()
        assert self._fetch_events_thread is not None
        _logger.info("[%s] Fetching timestamped events from thread in mode %r", self._name, self._mode)
        return self._fetch_events_thread.get_events(self.MAX_EVENTS_PER_CALL)

    @rpc_method
    def get_resolution(self) -> float:
        """Get time resolution.

        Returns:
            The resolution at the current binning (histogram bin width) as a float, in [ps]. Not meaningful in T2 mode.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            resolution = ctypes.c_double()
            self._lib.GetResolution(self._devidx, resolution)
            return resolution.value

    @rpc_method
    def get_sync_rate(self) -> int:
        """Return SYNC rate in events per second (determined every 100 ms).

        Allow at least 100 ms after :func:`initialize` or :func:`setSyncDivider` to get a stable rate meter reading.
        Similarly, wait at least 100 ms to get a new reading (100 ms is the gate time of the counter).

        Returns:
            Sync event rate in events per second.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            syncrate = ctypes.c_int()
            self._lib.GetSyncRate(self._devidx, syncrate)
            return syncrate.value

    @rpc_method
    def get_count_rate(self, channel: int) -> int:
        """Return input channel rate in events per second (determined every 100 ms).

        Allow at least 100 ms after :func:`initialize` to get a stable rate meter reading.
        Similarly, wait at least 100 ms to get a new reading (100 ms is the gate time of the counter).

        Parameters:
            channel: Channel index (range from 0 to `number_of_channels` - 1).

        Returns:
            Channel event rate in events per second.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            cntrate = ctypes.c_int()
            self._lib.GetCountRate(self._devidx, channel, cntrate)
            return cntrate.value

    @rpc_method
    def get_elapsed_measurement_time(self) -> float:
        """Get elapsed measurement time.

        Obtain the elapsed measurement time of a measurement. This relates to the current measurement when
        still running or to the previous measurement when already finished.

        Returns:
            The elapsed measurement time, in [ms].

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            elapsed = ctypes.c_double()
            self._lib.GetElapsedMeasTime(self._devidx, elapsed)
            return float(elapsed.value)

    @rpc_method
    def get_sync_period(self) -> float:
        """Get the sync period.

        This method only gives meaningful results while a measurement is running and after two sync periods have
        elapsed. The return value is undefined in all other cases. Resolution is that of the device's base
        resolution. Accuracy is determined by single shot jitter and clock stability.

        Returns:
            The sync period in [s].

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            period = ctypes.c_double()
            self._lib.GetSyncPeriod(self._devidx, period)
            return period.value

    @rpc_method
    def set_block_events(self, blocked: bool) -> None:
        """Enable or disable blocking event data.

        This feature can be used to "pause" the event data stream.
        The instrument will continue measuring and tagging events, but all
        event records will be discarded by the QMI driver.

        While events are blocked, the instrument will maintain its time base
        such that time intervals between events before and after the pause
        will be correctly represented.

        If real-time histograms are enabled, these will still be updated
        while events are blocked.

        This function only affects measurements in T2 mode.

        Parameter:
            block: True to block all event data, False to resume reporting event data.
        """
        self._check_is_open()
        assert self._fetch_events_thread is not None
        _logger.info("[%s] Setting event data stream blocking to %r", self._name, blocked)
        self._fetch_events_thread.set_block_events(blocked)

    @rpc_method
    def set_event_filter(self,
                         reset_filter: bool = False,
                         channel_filter: Optional[Dict[int, EventFilterMode]] = None,
                         sync_aperture: Optional[Tuple[int, int]] = None,
                         ) -> None:
        """Configure the event filter.

        When measuring in T2 mode, incoming events are passed through a filter
        to reject uninteresting events. This event filter is implemented in
        the QMI driver software (not inside the instrument).

        The filter is configured separately for each event type.
        Event types 0 to 7 correspond to the normal input channels.
        Event type 64 represents SYNC events.

        Each event type can be configured either to reject all events (`EventFilterMode.NO_EVENTS`),
        or to accept all events (`EventFilterMode.ALL_EVENTS`) or to accept only events that occur
        within a specific time window after a SYNC event (`EventFilterMode.APERTURE`).

        If `EventFilterMode.APERTURE` is selected for the SYNC channel, the filter accepts
        only the SYNC events that are either preceded or followed by a non-SYNC events.
        This mode can be used to discard redundant SYNCs when nothing interesting is happening.

        Parameters:
            reset_filter:   Configure the filter so that all normal events and SYNC events are accepted.
            channel_filter: Mapping from event type to the new event filter mode to be configured.
                            Any event types not specified in the mapping are left in their current setting.
            sync_aperture:  Tuple (delta_min, delta_max) in multiples of the instrument base resolution,
                            defining a time window following each SYNC event.
                            If unspecified or None, the current aperture setting is left unchanged.
        """
        self._check_is_open()
        assert self._fetch_events_thread is not None

        if reset_filter:
            # Reset filter so that all event types are accepted (normal input channels, SYNC).
            self._event_filter_channels.clear()
            for event_type in range(NUM_CHANNELS):
                self._event_filter_channels[event_type] = EventFilterMode.ALL_EVENTS
            self._event_filter_channels[SYNC_TYPE] = EventFilterMode.ALL_EVENTS
            _logger.info("[%s] Reset event filter to accept all event types", self._name)

        if channel_filter is not None:
            self._event_filter_channels.update(channel_filter)
            _logger.info("[%s] Set event filter to filter channels %s", self._name, channel_filter)

        if sync_aperture is not None:
            self._event_filter_aperture = sync_aperture
            _logger.info("[%s] Set event sync aperture to be %s", self._name, sync_aperture)

        self._fetch_events_thread.set_event_filter_config(self._event_filter_channels, self._event_filter_aperture)

    @rpc_method
    def set_realtime_histogram(self, channels: List[int], bin_resolution: int, num_bins: int, num_sync: int) -> None:
        """Configure real-time histograms.

        When measuring in T2 mode, the driver can optionally report real-time histograms.
        Each received event is assigned to a histogram bin based on the time interval between
        the last SYNC and the event. The histogram is integrated during several SYNC periods.
        After a configurable number of SYNC events, the integrated histogram is published
        via `sig_histogram`.

        Parameters:
            channels:       List of channels to include in the histogram (range 0 to 7).
            bin_resolution: Resolution of each histogram bin as a multiple of the instrument base resolution.
            num_bins:       Number of bins in the histogram (determines the maximum time interval after SYNC).
            num_sync:       Number of SYNC periods to integrate before publishing the histogram.
                            Specify 0 to disable real-time histograms.
        """
        if bin_resolution < 1:
            raise ValueError("Invalid bin_resolution")
        self._check_is_open()
        assert self._fetch_events_thread is not None
        self._fetch_events_thread.set_histogram_config(channels, bin_resolution, num_bins, num_sync)
        _logger.info(
            "[%s] Set histogram for channels %s to have resolution of %i with %i bins and %i sync periods",
            self._name, channels, bin_resolution, num_bins, num_sync
        )

    @rpc_method
    def set_realtime_countrate(self, sync_aperture: Tuple[int, int], num_sync: int) -> None:
        """Configure real-time count rate reporting.

        When measuring in T2 mode, the driver can optionally report real-time count rates.
        Events that fall in a specific time window after the SYNC pulse are counted.
        This counter is integrated during several SYNC periods.
        After a configurable number of SYNC events, the integrated counter values are
        published via `sig_countrate`.

        Parameters:
            sync_aperture:  Tuple (delta_min, delta_max) as a multiple of the instrument base resolution,
                            defining a time window after each SYNC pulse.
            num_sync:       Number of SYNC periods to integrate before publishing the counter value.
                            Specify 0 to disable the real-time countrate monitoring.
        """
        self._check_is_open()
        assert self._fetch_events_thread is not None
        self._fetch_events_thread.set_countrate_config(sync_aperture, num_sync)
        _logger.info(
            "[%s] Set real-time countrate with sync aperture %s with %i sync periods",
            self._name, sync_aperture, num_sync
        )

    @rpc_method
    def set_histogram_length(self, lencode: int) -> int:
        """Set the length of the histogram.

        Set the number of bins of the collected histograms. The histogram length is 65536 which is also the
        default after initialization if `set_histogram_length` is not called.

        Parameters:
            lencode: The histogram length code requesting a histogram of (1024 * 2**`lencode`) bins.
                     Valid range is from 0 to 6.

        Returns:
            The current length (time bin count) of histograms calculated as (1024 * 2**`lencode`).

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            actuallen = ctypes.c_int()
            self._lib.SetHistoLen(self._devidx, lencode, actuallen)
            self._actuallen = actuallen.value  # Update actual histogram buffer size
            return actuallen.value  # returns the actual length

    @rpc_method
    def set_measurement_control(self, meascontrol_str: str, startedge_str: str, stopedge_str: str) -> None:
        """Set the measurement control mode.

        This must be called before starting a measurement. The default after initialization (if this function is not
        called) is 0, i.e., software controlled acquisition time. The other modes 1..5 allow hardware triggered
        measurements through TTL signals at the control port or through White Rabbit. Note that this needs custom
        software.

        Parameters:
            meascontrol_str: Measurement control code. Any of 'SINGLESHOT_CTC', 'C1_GATED', 'C1_START_CTC_STOP',
                'C1_START_C2_STOP', 'WR_M2S', 'WR_S2M'.
            startedge_str : Start edge selection. Either 'RISING' or 'FALLING'.
            stopedge_str: Stop edge selection. Either 'RISING' or 'FALLING'.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        raise NotImplementedError()

    @rpc_method
    def clear_histogram_memory(self) -> None:
        """Reset the histogram state to all-zeros.

        Clear the histogram memory of all channels. Only meaningful in histogramming mode.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.ClearHistMem(self._devidx)

    @rpc_method
    def set_offset(self, offset: int) -> None:
        """Set time offset.

        This offset only applies in histogramming and T3 mode. It affects only the difference between stop and start
        before it is put into the T3 record or is used to increment the corresponding histogram bin.

        It is intended for situations where the range of the histogram is not long enough to look at "late" data. By
        means of the offset the "window of view" is shifted to a later range.

        This is not the same as changing or compensating cable delays. If the latter is desired,
        use :func:`setSyncChannelOffset` and/or :func:`setInputChannelOffset`.

        Parameters:
            offset: time offset in [ns], ranging from 0 ns to 100000000 ns.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.SetOffset(self._devidx, offset)

    @rpc_method
    def set_binning(self, binning: int) -> None:
        """Set time binning.

        Parameters:
            binning: the following values can be used:

                     | 0 = 1× base resolution,
                     | 1 = 2× base resolution,
                     | 2 = 4× base resolution,
                     | 3 = 8× base resolution, and so on.

                     Range is 0 to (`BINSTEPSMAX` - 1).

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.SetBinning(self._devidx, binning)

    @rpc_method
    def set_stop_overflow(self, stop_ovfl: bool, stopcount: int) -> None:
        """Set stop overflow count value.

        This setting determines if a measurement run will stop if any channel reaches the maximum set by the
        `stopcount` parameter.

        If `stop_ovfl` is False, the measurement will continue but counts above `STOPCNTMAX` in any bin will be clipped.

        Parameters:
            stop_ovfl: If true, the measurement will stop once the limit is reached. If not, the measurement will
                continue.
            stopcount: 1 to 4294967295.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            stop_ovfl_int = 1 if stop_ovfl else 0
            self._lib.SetStopOverflow(self._devidx, stop_ovfl_int, stopcount)

    @rpc_method
    def set_sync_divider(self, divider: int) -> None:
        """Set the SYNC divider value.

        The sync divider must be used to keep the effective sync rate at values < 78 MHz.
        It should only be used with sync sources of stable period.

        Using a larger divider than strictly necessary does not do great harm but it may result in slightly larger
        timing jitter.

        The readings obtained with :func:`get_count_rate` are internally corrected for the divider setting and
        deliver the external (undivided) rate. The sync divider should not be changed while a measurement is running.

        Parameters:
            divider: value ranging from 1 to 16.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.SetSyncDiv(self._devidx, divider)

    @rpc_method
    def set_sync_cfd(self, level: int, zc: int) -> None:
        """Set Constant Fraction Discriminator (CFD) levels for SYNC input.

        Notes:
            The values are given as a positive numbers although the electrical signals are actually negative.

        Args:
            level: CFD discriminator level in millivolts (mV) range 0 to 1000 representing 0 to -1000 mV.
            zc: CFD zero cross level in millivolts (mV). range 0 to 40 mV representing 0 to 40 mV.
        """
        raise NotImplementedError()

    @rpc_method
    def set_input_cfd(self, channel: int, level: int, zc: int) -> None:
        """Set Constant Fraction Discriminator (CFD) levels for input channel.

        Args:
            channel: Channel index (range from 0 to `number_of_channels` - 1).
            level: CFD discriminator level in millivolts (mV) range 0 to 1000 representing 0 to -1000 mV.
            zc: CFD zero cross level in millivolts (mV). range 0 to 40 mV representing 0 to 40 mV.
        """
        raise NotImplementedError()

    @rpc_method
    def set_sync_offset(self, sync_offset: int) -> None:
        """Set SYNC offset.

        This function can replace an adjustable cable delay. A positive offset corresponds to inserting a cable in
        the sync input.

        Parameters:
            sync_offset: SYNC offset, in [ps]. A value ranging from -99999 ps to +99999 ps.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        raise NotImplementedError()

    @rpc_method
    def set_sync_channel_offset(self, value: int) -> None:
        """Set SYNC channel offset.

        This is equivalent to changing the cable delay on the sync input. Actual resolution is the device's base
        resolution.

        Parameters:
            value: SYNC channel offset, in [ps]. A value ranging from -99999 ps to +99999 ps.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        raise NotImplementedError()

    @rpc_method
    def set_input_channel_offset(self, channel: int, value: int) -> None:
        """Set input channel offset.

        This is equivalent to changing the cable delay on the chosen input. Actual resolution is the device's base
        resolution.

        Parameters:
            channel: Channel index (range from 0 to `number_of_channels` - 1).
            value: channel offset, in [ps]. A value ranging from -99999 ps to +99999 ps.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.SetInputChannelOffset(self._devidx, channel, value)

    @rpc_method
    def set_input_channel_enable(self, channel: int, enable: bool) -> None:
        """Enable or disable the input channel.

        Parameters:
            channel: Channel index (range from 0 to `number_of_channels` - 1).
            enable: True to enable the channel, False to disable.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            enable_int = 1 if enable else 0
            self._lib.SetInputChannelEnable(self._devidx, channel, enable_int)

    @rpc_method
    def get_flags(self) -> List[str]:
        """Get flags.

        Returns:
            A list of active flags, represented as short strings.
            See the definition of the :class:`~qmi.instruments.picoquant.<xxx>harp_wrapper.FLAG`
            type for possible values. <xxx> = multi, hydra, pico or time.
            The meaning of these flags is not fully documented.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        raise NotImplementedError()

    @rpc_method
    def get_histogram(self, channel: int, clear: Optional[int] = None) -> np.ndarray:
        """Get histogram data array from a specific channel. Note that MH_GetHistogram cannot be used with the
        shortest two histogram lengths of 1024 and 2048 bins. You need to use MH_GetAllHistograms in this case.

        Parameters:
            channel: Channel index (range from 0 to `number_of_channels` - 1).
            clear: Optional input for HydraHarp only: 0 = keep histogram in buffer, 1 = clear buffer

        Returns:
              Histogram data array for given input channel.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            # The histogram buffer size must correspond to the default or the value obtained through MH_SetHistoLen().
            hist_data = np.empty(self._actuallen, dtype=np.uint32)
            ctypes_hist_data = hist_data.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32))
            if clear is not None:
                self._lib.GetHistogram(self._devidx, ctypes_hist_data, channel, clear)

            else:
                self._lib.GetHistogram(self._devidx, ctypes_hist_data, channel)

            return hist_data.copy()

    @rpc_method
    def get_all_histograms(self) -> np.ndarray:
        """Get histogram data array from a specific channel. The multidimensional array receiving the data must be
         shaped according to the number of input channels of the device and the chosen histogram length.

        Returns:
              Histogram data arrays of all channels.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        raise NotImplementedError()

    @rpc_method
    def get_warnings(self) -> List[str]:
        """Get a list of warnings.

        Returns:
            A list of strings. See the definition of the :class:`~qmi.instruments.picoquant.<xxx>harp_wrapper.WARNING`
            type for possible values. <xxx> = multi, hydra, pico or time.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        raise NotImplementedError()

    @rpc_method
    def get_warnings_text(self, warnings: int) -> str:
        """Translates warnings into human-readable text.

        Parameters:
            warnings: integer bitfield obtained from MH_GetWarnings

        Returns:
            The warning text string.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            text = ctypes.create_string_buffer(16384)
            self._lib.GetWarningsText(self._devidx, text, warnings)
            return text.value.decode()

    @rpc_method
    def set_trigger_output(self, period: int) -> None:
        """Set trigger output.

        Set the period of the programmable trigger output. The period 0 switches it off.

        Parameters:
            period: Period in units of 100 ns range is 0 to 16777215.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        raise NotImplementedError()

    @rpc_method
    def get_module_info(self) -> List[Tuple[int, int]]:
        """Get the model and version codes of all modules of the device.

        Note:
            This function can only be used after the :func:`initialize` method was successfully called.

        Returns:
            List of tuples (modelcode, versioncode).
            This information may be needed for support queries.
            The meaning of these numbers is not documented.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        raise NotImplementedError()



import unittest
from unittest.mock import patch, Mock

from qmi.core.context import QMI_Context
from qmi.instruments.raspberry import RaspberryPi_Gpio


class RasperryPiGpioTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.patcher = patch("qmi.instruments.raspberry.gpio.GPIO")
        self.board = [3, 5, 7, 8, 10, 11, 12, 13, 15, 16, 18, 19, 21, 22, 23, 24, 26, 29, 31, 32, 33, 35, 36]
        self.high = 1
        self.low = 0
        self.patcher.start()
        self.patcher.target.GPIO.BOARD = self.board
        self.patcher.target.GPIO.HIGH = self.high
        self.patcher.target.GPIO.LOW = self.low
        self.gpio = RaspberryPi_Gpio(QMI_Context("rpi_test"), "rpi_gpio")

    def tearDown(self) -> None:
        self.patcher.stop()

    def test_open_close(self):
        """Test open and close functions."""
        self.patcher.target.GPIO.setmode = Mock()
        # with patch("qmi.instruments.raspberry.gpio.GPIO.setmode") as sm_patch:
        self.gpio.open()

        self.patcher.target.GPIO.setmode.assert_called_once_with(self.board)

        self.patcher.target.GPIO.cleanup = Mock()
        self.gpio.close()

        self.patcher.target.GPIO.cleanup.assert_called_once_with()

    def test_gpio_setup(self):
        """Test gpio_setup function with in oder aus."""
        # case IN
        self.gpio._is_open = True
        pin_nr = 15
        gpio_mode = self.gpio.gpio_mode.INPUT
        with patch("qmi.instruments.raspberry.gpio.GPIO.setup") as sp_patch:
            self.gpio.gpio_setup(pin_nr, gpio_mode)

        sp_patch.assert_called_once_with(pin_nr, self.patcher.target.GPIO.IN)

        # case OUT
        gpio_mode = self.gpio.gpio_mode.OUTPUT
        with patch("qmi.instruments.raspberry.gpio.GPIO.setup") as sp_patch:
            self.gpio.gpio_setup(pin_nr, gpio_mode)

        sp_patch.assert_called_once_with(pin_nr, self.patcher.target.GPIO.OUT)

    def test_gpio_input(self):
        """Test gpio_input function."""
        self.gpio._is_open = True
        pin_nr = 3
        self.patcher.target.GPIO.HIGH = self.high
        self.patcher.target.GPIO.input = Mock(return_value=1)

        success = self.gpio.gpio_input(pin_nr)

        self.assertTrue(success)
        self.patcher.target.GPIO.input.assert_called_once_with(pin_nr)

    def test_gpio_output(self):
        """Test gpio_output function."""
        self.gpio._is_open = True
        pin_nr = 34
        self.patcher.target.GPIO.HIGH = self.high
        self.patcher.target.GPIO.LOW = self.low
        self.patcher.target.GPIO.output = Mock()

        # case value  = True
        value = True
        self.gpio.gpio_output(pin_nr, value)

        self.patcher.target.GPIO.output.assert_called_once_with(pin_nr, self.high)

        self.patcher.target.GPIO.output.reset_mock()
        # case False
        value = False
        self.gpio.gpio_output(pin_nr, value = False)

        self.patcher.target.GPIO.output.assert_called_once_with(pin_nr, self.low)


if __name__ == '__main__':
    unittest.main()
